#!perl -w

use strict;
use Cwd;

my @contour_intervals = (25, 5, 1); #, 1.25, 0.625);
my ($laz, $dxf);

# tile_442800_6620400.laz

my $FORCE = 0;
my $CORES = 0;
my $DEPRESSIONS = 0;
my $SKIPZERO = 0;

my $INNER_LOOPS = 3;

my @args = (); # Extra args for las2iso!

while ($_ = shift) {
	if (/^-ci$/i || /^-contour_intervals$/i) {
		$_ = shift;
		@contour_intervals = split(/,/);
	}
	elsif (/^-force$/i) {
		$FORCE = 1;
	}
	elsif (/^-depressions$/i) {
		$DEPRESSIONS = 1;
	}
	elsif (/^-skipzero$/i) {
		$SKIPZERO = 1;
	}
	elsif (/^-cores$/i) {
		$CORES = shift;
		if (! ($CORES =~ /^\d+$/)) {
			$CORES = 0;
		}
	}
	elsif (/^-arg$/i) {
		push(@args, shift);
	}
	elsif (!defined($laz)) {
		$laz = $_;
	}
	elsif (!defined($dxf)) {
		$dxf = $_;
	}
}

if ($contour_intervals[-1] <= 1) {
#	$INNER_LOOPS = 2;
}

my $LIMIT = 1250000;

# Pick up current directory name, use as default output base:
my ($TXT, $ASC) = ('','');
if (! defined($dxf)) {
	$dxf = getcwd;
	if ($dxf =~ /^(.+)[\\\/]([^\\\/]+)$/) {
		$dxf = $2;
	}
	$TXT = $dxf . '_contours.txt';
	$ASC = $dxf . '_elevation.asc';
	$dxf = $dxf . '_contours.dxf';
}
else {
	$dxf = $dxf . '.dxf' unless ($dxf =~ /\.dxf$/);
	$TXT = substr($dxf,0,length($dxf)-4) . '_contours.txt';
	$ASC = substr($dxf,0,length($dxf)-4) . '_elevation.asc';
}

my @files = myglob($laz);

if (scalar(@files) == 1 && ! ($files[0] =~ /\.la[sz]$/i)) { # single non-LA[SZ] file, use as List of Files:
	open(LOF, $files[0]) || die($files[0].' is not a list of files!\n');
	@files = ();
	while (<LOF>) {
		chomp;
		push(@files, $_);
	}
	close(LOF);
}

die("Usage: gen-dxf-iso.pl lazfile|file_with_list_of_laz_files [output_file]\n") unless (scalar(@files) && scalar(@contour_intervals));

my ($XSCALE, $YSCALE, $HSCALE) = (10,10,100);  # Room for 0.25 m contours
my ($XINV, $YINV, $HINV) = (1/$XSCALE, 1/$YSCALE, 1/$HSCALE);

my @contour_layers = ('102_index_contour', '101_contour', '103_form_line');
my %crt = ( '102_index_contour' => 102, '101_contour' => 101, '103_form_line' => 103,
	'104_slope_line' => 104, '115_depression' => 115, '112_dot_knoll' => 112, 
	'101.999_negative_contour' => 101.999,
	'102.999_negative_index_contour' => 102.999,
	'103.999_negative_form_line' => 103.999);
	
my %standard_contour = ('102_index_contour' => 1, '101_contour' => 2, '103_form_line' => 3);

# Do we have any additional form line layers? If so, add them and give them a layer name:
my ($form, $neg_form) = (103.000, 103.999);
for (my $i = 3; $i < scalar(@contour_intervals); $i++) {
	$form += 0.001;
	$neg_form -= 0.001;
	my $layer_name = sprintf('%7.3f_form_line_%d', $form, $i-2);
	push(@contour_layers, $layer_name);
	$crt{$layer_name} = sprintf('%7.3f', $form);
	$layer_name = sprintf('%7.3f_negative_form_line_%d', $neg_form, $i-2);
	push(@contour_layers, $layer_name);
	$crt{$layer_name} = sprintf('%7.3f', $neg_form);
}

my %height2layer = ();

# height2layer receives heights in integer cm and returns the proper layer name
sub height2layer
{
	my ($height) = @_;
	#my $height = sprintf("%1.0f", $h * $HSCALE);
	if (defined($height2layer{$height})) {
		return $height2layer{$height};
	}

	my $h = abs($height);
	# Find contour type by starting at the index contours and going down until
	# we find one which is a true divisor of the current height:
	for (my $i = 0; $i < scalar(@contour_intervals); $i++) {
		if (($h % ($contour_intervals[$i]*$HSCALE)) == 0) {
			my $layer = $contour_layers[$i];
			$height2layer{$height} = $layer;
			if ($height < 0) {
				my $crt = $crt{$layer};
				if ($crt == 102 || $crt == 101) {
					$crt = 101.999;
					$layer = '101.999_negative_contour';
				}
				else { # $crt >= 103
					my $frac = $crt - int($crt);
					$crt = 103.999 - $frac;
					$layer = sprintf("%7.3f_negative_form_line", $crt);
					$crt{$layer} = $crt;
				}
			}
				
			return $layer;
		}
	}
	# If this was an intermediate layer which should be skipped, return an empty layer name:
	# Remember this for any subsequent contours of the same intermediate height
	$height2layer{$height} = '';
	return '';
}

my $contour_interval = 0.25; # gcd(@contour_intervals);

my ($XMIN, $XMAX, $YMIN, $YMAX, $STEP) = find_min_max(@files);

my ($tid, $core_merge, @tids, @core_merge) = (1,'');

my $CID = '';
if ($CORES > 1) { # Start sub-processes!
	my $step = int((scalar(@files)+$CORES-1) / $CORES);
	my $beg = 0;
	for (my $core = 0; $core < $CORES; $core++, $beg += $step) {
		my @sub_tiles = ();
		my $end = $beg + $step;
		$end = scalar(@files) if ($end > scalar(@files));
		for (my $f = $beg; $f < $end; $f++) {
			push(@sub_tiles, $files[$f]);
		}
		$core_merge = sprintf('%s_%d.txt', substr($dxf,0,length($dxf)-4), $core);
		if (! -f $core_merge || $FORCE) {
			$tid = fork;
			if ($tid == 0) {
				@tids = ();
				@files = @sub_tiles;
				$CID = "$core: ";
				last;
			}
			printf(STDERR "PID $tid forked\n");
			push(@tids, $tid);
		}
		push(@core_merge, $core_merge);
	}
}

my @contours = ();
my @slopes = ();
my @dep = ();
my @knolls = ();

if ($tid == 0 || ((scalar(@tids) == 0) && (scalar(@core_merge) < 2))) { # Single-thread processing
	if (!$FORCE && $tid == 0 && -f $core_merge) {
		die("$CID $core_merge already exists, exiting!\n");
	}
		
	my $fnr = 0;
	foreach (@files) {
		chomp;
		printf(STDERR "%s%d/%d %s\n", $CID, ++$fnr, scalar(@files), $_);
		my ($xmin, $xmax, $ymin, $ymax, $dep, $knoll, $slope, @c);
		if (defined($STEP) && /tile_([\-]?\d+)_([\-]?\d+)\D/) {
			my ($e, $n, $s) = ($1, $2, $STEP);
			if (/tile_([\-]?\d+)_([\-]?\d+)_(\d+)\D/) {
				($e, $n, $s) = ($1, $2, $3);
			}
			($xmin, $xmax, $ymin, $ymax, $dep, $knoll, $slope, @c) = gen_tile_contours($_, $contour_interval, $1, $1+$s, $2, $2+$s);
		}
		else {
			($xmin, $xmax, $ymin, $ymax, $dep, $knoll, $slope, @c) = gen_tile_contours($_, $contour_interval);
		}
		push(@contours, @c);
		push(@slopes, @$slope);
		push(@knolls, @$knoll);
		push(@dep, @$dep);
	}

	@contours = merge_contours(@contours) if (scalar(@files) > 1);
	
	if ($tid == 0) { # Is this a sub-thread?
		write_txt($core_merge, $XMIN, $XMAX, $YMIN, $YMAX, \@dep, \@knolls, \@slopes, \@contours);
		die("$CID Done\n");
	}
}
else { # Parent process!
	my $kid;
	foreach (@tids) {
		$kid = waitpid($_, 0);
		#sleep(1);
	}
	printf(STDERR "Finished waiting, collecting raw tiles\n");

	foreach(@core_merge) {
		printf(STDERR "Loading %s\n", $_);
		open(my $fh, '<', $_);
		my ($dep, $knoll, $slope, $cont) = load_txt($fh);
		push(@dep, @$dep);
		push(@knolls, @$knoll);
		push(@slopes, @$slope);
		push(@contours, @$cont);
		close($fh);
	}
	@contours = merge_contours(@contours);
}

#printf(STDERR "Writing final %s file\n", $TXT);
#write_txt($TXT, $XMIN, $XMAX, $YMIN, $YMAX, \@dep, \@knolls, \@slopes, \@contours);

if ($DEPRESSIONS && -f $ASC) {
	my ($point_depressions, $slope_lines);
	printf(STDERR "Looking for depressions...");
	($point_depressions, $slope_lines, @contours) = find_depressions($ASC, @contours);
	push(@dep, @$point_depressions);
	push(@slopes, @$slope_lines);
	printf(STDERR " found %d points depressions and placed %d fall line tags\n", 
		scalar(@$point_depressions), scalar(@$slope_lines));
}

printf(STDERR "Writing final %s file\n", $dxf);
write_dxf($dxf, $XMIN, $XMAX, $YMIN, $YMAX, \@dep, \@knolls, \@slopes, @contours);

#create crt file for OCAD import:
my $crt = substr($dxf,0,length($dxf)-3)."crt";
open(CRT, ">$crt") || die("Could not create '$crt' file\n");
foreach (sort {$crt{$a} <=> $crt{$b}} keys %crt) {
	printf(CRT "%s %s\n", $crt{$_}, $_);
}
close(CRT);
printf(STDERR "Done! CRT file for OCAD import: %s\n", $crt);

sub gen_tile_contours
{
	my ($fname, $interval, $xmin, $xmax, $ymin, $ymax) = @_;
	
	# Convert to backwards directory slashes for lastools binaries:
	$fname =~ s/\//\\/g;
	
	($xmin, $xmax, $ymin, $ymax) = find_min_max($fname) unless defined($ymax);

	my $base = substr($fname,0,length($fname)-4);
	#printf(STDERR "%s%s X: %1.2f - %1.2f\nY: %1.2f - %1.2f\n", $CID, $fname, $xmin, $xmax, $ymin, $ymax);
	my $txtname = $base.'_iso.txt';

	# Generate contours:
	my (@c, @dep, @knolls, @slope) = ();
	if (!$FORCE && -e $txtname) {
		printf(STDERR "%s$txtname exists\n", $CID);
		if ($tid) {
			open(my $fh, '<', $txtname) || die("Could not read from $txtname!");
			#@c = load_txt($fh, $xmin, $xmax, $ymin, $ymax);
			my ($dep, $knoll, $slope, $cont) = load_txt($fh, $xmin, $xmax, $ymin, $ymax);
			push(@dep, @$dep);
			push(@knolls, @$knoll);
			push(@slopes, @$slope);
			push(@c, @$cont);
		}
	}
	else {
		
#		my $cmd = sprintf('las2iso -i %s -otxt -stdout -keep_class 2 -smooth 2 -simplify 1 -concavity 0 -iso_every %s 2>nul', 
#			$fname, $interval);
		my $cmd = sprintf('las2iso -i %s -otxt -stdout -keep_class 2 -simplify 1 -concavity 20 -iso_every %s 2>nul', 
			$fname, join(" ", $interval, @args));
		printf(STDERR "%s%s\n", $CID, $cmd);
		open(my $fh, "-|", $cmd) || die ("$CID Could not open pipe from las2iso!\n");
		my ($dep, $knoll, $slope, $cont) = load_txt($fh);
		@c = @$cont;
		# Clip off 5 m from the edge to get rid of artifacts!
		@c = clip_contours($xmin-30, $xmax+30, $ymin-30, $ymax+30, @c);
		
		# Look for knolls: Must have at least three small concentric rings above/below each other
		# Ditto for point depressions: Max size = 20x20 m
		
		($dep, $knoll, $slope, @c) = find_knolls_and_depressions($xmin, $xmax, $ymin, $ymax, @c);
		push(@dep, @$dep);
		push(@knolls, @$knoll);
		push(@slope, @$slope);
		
		my $pre = scalar(@c);
		@c = clip_contours($xmin, $xmax, $ymin, $ymax, @c);
		printf(STDERR "%s%d contours, %d after filtering\n", $CID, $pre, scalar(@c));
		#write_txt($txtname, $xmin, $xmax, $ymin, $ymax, \@dep, \@knolls, \@slope, \@c) if ($tid);
	}

	return ($xmin, $xmax, $ymin, $ymax, \@dep, \@knolls, \@slope, @c);
}

sub fix_height
{
	my ($h, $cont) = @_;
	my $oh = $h; $h = abs($h);
	if ($oh >= 0) {
		# Find nearest regular contour above:
		while (!$standard_contour{height2layer($h)}) {
			$h += 25;
		}
	}
	else { # Negative contour, go down!
		# Ditto below
		while (!$standard_contour{height2layer($h)}) {
			$h -= 25;
		}
	}
	$h = -$h if ($oh < 0);
	my @points = ();
	foreach (split(/\n/, $cont)) {
		my ($x, $y, $hd) = split(/,/);
		push(@points, join(",", $x, $y, $h));
	}
	$cont = join("\n", @points);
	return ($h, $cont);
}

sub find_knolls_and_depressions
{
	my ($xmin, $xmax, $ymin, $ymax, @contours) = @_;
	
	my @idx = ();
	my (@knolls, @dep, @slope, @c) = ();
	my %depression_line = ();
	my %knoll_line = ();
	my %negative_contour = ();
	#printf(STDERR "Generating index from %d raw contours\n", scalar(@contours));
	for (my $c = 0; $c < scalar(@contours); $c++) {
		my @points = split(/\n/, $contours[$c]);
		next unless (key($points[0]) eq key($points[-1])); # Not a Loop, so skip!
		
		my ($x0, $x1, $y0, $y1) = (1e38,-1e38,1e38,-1e38);
		my ($x, $y, $h);
		foreach (@points) {
			($x, $y, $h) = split(/,/);
			$x0 = $x if ($x < $x0);
			$x1 = $x if ($x > $x1);
			$y0 = $y if ($y < $y0);
			$y1 = $y if ($y > $y1);
		}
		push(@idx, join(',', $x0,$x1,$y0,$y1,$h,$c));
	}
	@idx = sort(@idx);
	#printf(STDERR "Found %d loops, looking for internal contours\n", scalar(@idx));
	for (my $i = 0; $i < scalar(@idx); $i++) {
		my ($x0,$x1,$y0,$y1,$h,$c) = split(/,/, $idx[$i]);
		my ($cb, $db, $hb, $dj);
		for (my $j = $i+1; $j < scalar(@idx); $j++) {
			my ($x0j,$x1j,$y0j,$y1j,$hj,$cj) = split(/,/, $idx[$j]);
			last if ($x0j > $x1);
			
			next if ($x1j > $x1 || $y0j < $y0 || $y1j > $y1);
			$dj = $x0j-$x0 + $x1-$x1j + $y0j-$y0 + $y1-$y1j;
			if (!defined($db) || $dj < $db) {
				$cb = $j;
				$hb = $hj;
				$db = $dj;
			}
		}
		if (defined($cb) && ($dj*$XINV < 100)) { # We have found the nearest internal ring and it is not too far away!
			$idx[$i] = join(',', $x0,$x1,$y0,$y1,$h,$c, $cb);
		}
	}
	#printf(STDERR "Checking for third innermost loops\n");
	my $cnt = 0;
	for (my $i = 0; $i < scalar(@idx); $i++) {
		my ($x0,$x1,$y0,$y1,$h,$c, $b) = split(/,/, $idx[$i]);
		next unless defined($b);
		
		my ($x0i,$x1i,$y0i,$y1i,$hi,$ci, $bi) = split(/,/, $idx[$b]);
		if ($hi < $h) {
			$negative_contour{$c} = 1;
		}
		my ($x0j,$x1j,$y0j,$y1j,$hj,$cj, $bj);
		
		if ($INNER_LOOPS == 2) {
			next if (defined($bi));
			#$cj = $ci;
			($x0j,$x1j,$y0j,$y1j,$hj,$cj, $bj) = ($x0i,$x1i,$y0i,$y1i,$hi,$ci, $bi);
		}
		else {
			next unless defined($bi);
		
			($x0j,$x1j,$y0j,$y1j,$hj,$cj, $bj) = split(/,/, $idx[$bi]);
			($x0i,$x1i,$y0i,$y1i,$hi,$ci, $bi) = ($x0j,$x1j,$y0j,$y1j,$hj,$cj, $bj);
			next if defined($bj);
		}
		
		$cnt++;
		
		# Check area:
		my $a = ($x1-$x0 + $y1-$y0) * $XINV;
		if ($hi < $h) { # Depression, check size:
			if ($a > 40) { # Large enough (20x20) for help line contour depression
				# Adjust the height to the closest regular (or help) contour
				($h, $contours[$c]) = fix_height(-$h, $contours[$c]);
				
				# Locate extremal points for the slope tag(s), keeping the outer loop
				my ($x, $y, $h, $x0y, $x1y, $y0x, $y1x);
				($x0, $x1, $y0, $y1) = (1e38,-1e38,1e38,-1e38);
				my @points = split(/\n/, $contours[$c]);
				next unless scalar(@points);
				foreach (@points) {
					($x, $y, $h) = split(/,/);
					if ($x < $x0) {
						$x0 = $x;
						$x0y = $y;
					}
					if ($x > $x1) {
						$x1 = $x;
						$x1y = $y;
					}
					if ($y < $y0) {
						$y0x = $x;
						$y0 = $y;
					}
					if ($y > $y1) {
						$y1x = $x;
						$y1 = $y;
					}
				}
				# Make sure that this contour stays on the map!
				$depression_line{$c} = 1;

				# Add slope tags:
				# (0, 90, 180, -90) is (up, left, down, right)
				if (! (defined($x0) && defined($x1) && defined($y0) && defined($y1) &&
					defined($x0y) && defined($x1y) && defined($y0x) && defined($y1x))) {
					die("Undefined?\n");
				}
				if ($x1-$x0 > $y1-$y0) { # -135 < dir < -45
					my $dir = atan2($x1y-$x0y, $x1-$x0) * 180/3.1415926535 - 90;
					push(@slope, join(',', $x0, $x0y, $dir));
					if ($x1-$x0 > 21*$XSCALE) {
						$dir += 180;
						$dir -= 360 if ($dir > 180);
						push(@slope, join(',', $x1, $x1y, $dir));
					}
				}
				else {
					my $dir = atan2($y1-$y0, $y1x-$y0x) * 180/3.1415926535 - 90;
					$dir += 360 if ($dir < -180);
					push(@slope, join(',', $y0x, $y0, $dir));
					if ($y1-$y0 > 21*$YSCALE) {
						$dir += 180;
						$dir -= 360 if ($dir > 180);
						push(@slope, join(',', $y1x, $y1, $dir));
					}
				}
				$contours[$ci] = $contours[$cj] = '';
			}
			else {
				my ($x, $y) = (($x0i+$x1i)*0.5*$XINV, ($y0i+$y1i)*0.5*$YINV);
				# Clip boundaries:
				next if ($x < $xmin || $x > $xmax || $y < $ymin || $y > $ymax);
				push(@dep, join(',', $x*$XSCALE, $y*$YSCALE, 0));
				#printf(STDERR "Found depression\n");
				$contours[$c] = $contours[$ci] = $contours[$cj] = '';
			}
		}
		else { # Knoll!
			if ($a > 30) { # Keep as regular contour, can be smaller than depression (no slope tags!)
				#($h, $contours[$c]) = fix_height($h, $contours[$c]);
				next;
			}
			
			my ($x, $y) = (($x0i+$x1i)*0.5*$XINV, ($y0i+$y1i)*0.5*$YINV);
			# Clip boundaries:
			next if ($x < $xmin || $x > $xmax || $y < $ymin || $y > $ymax);

			push(@knolls, join(',', $x*$XSCALE, $y*$YSCALE, $hi));
			#printf(STDERR "Found dot knoll\n");
			$contours[$c] = $contours[$ci] = $contours[$cj] = '';
		}
	}
	#printf(STDERR "%d candidates found\nFiltering %d contours, removing extra layers\n", 
	#	$cnt, scalar(@contours));
	@c = ();
	for (my $c = 0; $c < scalar(@contours); $c++) {
		$_ = $contours[$c];
		next if ($_ eq '');
		
		my @points = split(/\n/);
		my ($x,$y,$h) = split(/,/, $points[0]);
		my $layer = height2layer($h);
		next if ($layer eq '');
		if ($negative_contour{$c}) {
			$points[0] = join(',', $x,$y,-$h);
			$_ = $contours[$c] = join("\n", @points);
		}
		
		push(@c, $_);
	}
	printf(STDERR "%s%d contours (from %d), %d knolls, %d depressions, %d slope tags\n", 
		$CID, scalar(@c), scalar(@contours), scalar(@knolls), scalar(@dep), scalar(@slope));
	return (\@dep, \@knolls, \@slope, @c);
}

sub deltas
{
	my @steps = sort {$a <=> $b} @_;
	return (undef, undef) if (scalar(@steps) <= 1);

	my ($min, $max) = (1e38, 0);
	for (my $i = 1; $i < scalar(@steps); $i++) {
		my $step = $steps[$i]-$steps[$i-1];
		$min = $step if ($step < $min);
		$max = $step if ($step > $max);
	}
	return ($min, $max);
}

sub find_min_max
{
	my @files = @_;
	
	my ($xmin, $xmax, $ymin, $ymax, $adaptive) = (1e38,-1e38,1e38,-1e38, 0);
	if (scalar(@files) > 1) {
		# New & even faster method: tile_EEEEE_NNNNN_SSS.laz naming for adaptive tile sizes:
		
		my ($smin, $smax) = (1e38, 0);
		foreach (@files) {
			goto skip unless (/tile_([\-]?\d+)_([\-]?\d+)_([\d+)\D/);

			my ($e, $n, $s) = ($1, $2, $3);
			$xmin = $e if ($e < $xmin);
			$xmax = $e if ($e > $xmax);
			$ymin = $n if ($n < $ymin);
			$ymax = $n if ($n > $ymax);
		}
		$adaptive = -1;
		return ($xmin, $xmax, $ymin, $ymax, $adaptive);
		
skip:
		# Try to use fast method: Look for tile_EEEEEE_NNNNNNN.laz names!
		my %ehist = ();
		my %nhist = ();
		my $fast = 1;
		foreach (@files) {
			unless (/tile_([\-]?\d+)_([\-]?\d+)\D/) {
				$fast = 0;
				last;
			}
			$ehist{$1}++;
			$nhist{$2}++;
		}
		if ($fast) {
			my @ehist = sort {$a <=> $b} keys %ehist;
			my @nhist = sort {$a <=> $b} keys %nhist;
			my ($min, $max) = deltas(@ehist);
			my $step = $min;
			($min, $max) = deltas(@nhist);
			if ($min == $step) {
				$step = 250 if ($step < 250);
				
				$xmin = $ehist[0];
				$xmax = $ehist[-1] + $step;
				$ymin = $nhist[0];
				$ymax = $nhist[-1] + $step;
					
				return ($xmin, $xmax, $ymin, $ymax, $step);
			}
		}
	}
	my $ground_points = 0;
	foreach (@files) {
		chomp;
		my $path = $_;
		$path =~ s/\//\\/g;
		my $res = `lasinfo -stdout $path`;
		if (0) { #($res =~ /LAStiling .+ bbox (\S+)\s(\S+)\s(\S+)\s+(\S+), buffer/) {
			$xmin = $1 if ($1 < $xmin);
			$ymin = $2 if ($2 < $ymin);
			$xmax = $3 if ($3 > $xmax);
			$ymax = $4 if ($4 > $ymax);
		}
		else {
			die("Not a LAS/LAZ file? $_\n") unless ($res =~ /min x y z:\s+(\S+)\s+(\S+)\s+(\S+)/);
			$xmin = $1 if ($1 < $xmin);
			$ymin = $2 if ($2 < $ymin);
			die("Not a LAS/LAZ file? $_\n") unless ($res =~ /max x y z:\s+(\S+)\s+(\S+)\s+(\S+)/);
			$xmax = $1 if ($1 > $xmax);
			$ymax = $2 if ($2 > $ymax);
		}
		my $g = 0;
		if ($res =~ /\s(\d+)\s+Ground (2)/) {
			$g = $1;
		}
		$ground_points += $g;
	}
	#printf(STDERR "Raw X: %1.2f - %1.2f\nY: %1.2f - %1.2f\n", $xmin, $xmax, $ymin, $ymax);
	
	$xmin = adjust($xmin, 50);
	$ymin = adjust($ymin, 50);
	$xmax = adjust($xmax, -50);
	$ymax = adjust($ymax, -50);
	
	return ($xmin, $xmax, $ymin, $ymax);
}

sub adjust
{
	my ($x, $adj) = @_;
	my $ax = sprintf("%1.0f", $x+$adj);
	$x = $ax if (($ax % 1000) == 0);
	return $x;
}

sub load_txt
{
	my ($fh) = @_;
	
	# grab all contours, convert all coordinates to (rounded) mm
	my (@dep, @knoll, @slope, @contours) = ();
	my $save = $/;
	$/ = "\n#\n";
	while (<$fh>) {
		chomp;
		my @points = split(/\n/);
		if ($#points == 0) { # A single scalar, must be a point detail!
			my ($x, $y, $dir) = split(',', $points[0]);
			next unless defined($dir);
			
			my ($symbol, $symname, $val) = split('_', $dir);
			
			#my $symbol = int($dir / 360); $dir -= $symbol*360;
			#while ($dir < -180) {
			#	$dir += 360;
			#	$symbol--;
			#}
			#while ($dir > 180) {
			#	$dir -= 360;
			#	$symbol++;
			#}
			
			#my $point = sprintf("%1.0f,%1.0f,%s", $x*$XSCALE, $y*$YSCALE, $dir);
			
			if ($symbol == 104) { # Slope line
				push(@slope, sprintf("%1.0f,%1.0f,%s", $x*$XSCALE, $y*$YSCALE, $val));
			}
			elsif  ($symbol == 112) { # Dot knoll
				push(@knoll, sprintf("%1.0f,%1.0f,%s", $x*$XSCALE, $y*$YSCALE, $val));
			}
			elsif  ($symbol == 115) { # Depression
				push(@dep, sprintf("%1.0f,%1.0f", $x*$XSCALE, $y*$YSCALE));
			}
			next;
		}
		my @p = ();
		my ($x, $y, $h);
		foreach (@points) {
			($x, $y, $h) = split(/,/);
			next unless defined($h);
			
			push(@p, sprintf("%1.0f,%1.0f,%1.0f", $x*$XSCALE, $y*$YSCALE, $h*$HSCALE));
		}
		push(@contours, join("\n", @p));
	}
	close($fh);
	$/ = $save;

	return (\@dep, \@knoll, \@slope, \@contours);
}

sub is_loop
{
	my @points = @_;
	return key($points[0]) eq key($points[$#points]);
}

sub poly_len
{
	my @points = @_;
	my ($px, $py);
	my $len = 0;
	foreach (@points) {
		my ($x, $y, $h) = split(',');
		if (defined($px)) {
			my ($dx, $dy) = ($x-$px, $y-$py);
			$len += sqrt($x*$x+$y*$y);
		}
	}
	$len *= $XINV;	
	return $len;
}


sub write_dxf
{
	my ($dxf, $xmin, $xmax, $ymin, $ymax, $point_depressions, $dot_knolls, $slope_lines, @contours) = @_;
	
	# Intervals in integer steps!
	#my $interval = sprintf("%1.0f", $contour_interval*$HSCALE);

	open(DXF, ">$dxf") || die("Could not write to $dxf\n");
	
	printf(DXF "0\nSECTION\n2\nHEADER\n9\n\$EXTMIN\n10\n%1.2f\n20\n%1.2f\n9\n\$EXTMAX\n10\n%1.2f\n20\n%1.2f\n0\nENDSEC\n",
		$xmin, $ymin, $xmax, $ymax);
		
	printf(DXF "0\nSECTION\n2\nTABLES\n0\nTABLE\n2\nLAYER\n0\nLAYER\n2\nform_line\n0\nLAYER\n2\ncontour\n0\nLAYER\n2\nindex_contour\n0\nENDTAB\n0\nENDSEC\n");
	printf(DXF "0\nSECTION\n2\nENTITIES\n");

	my $POLY = 0;
	
	#my %layer_count = ();
	foreach (@contours) {
		my @points = split(/\n/);
		my $h = (split(/,/, $points[0]))[2];
		my $height = sprintf("%1.2f", abs($h)*$HINV);
		
		if (scalar(@points) < 50 && is_loop(@points) && poly_len(@points) < 20) {
			next; # Skip very short loops, they should be shown as dot knolls or depressions instead!
		}

		if ($SKIPZERO && $h == 0) {
			next;
		}
		my $layer = height2layer(abs($h));
		#die("Missing layer for $height\n") if ($layer eq '');
		if ($h < 0) {
			if ($layer =~ /^(\d\d\d)_/) {
				$layer = $1.'.999_negative'.substr($layer,3);
			}
			elsif ($layer =~ /^(\d\d\d)\.(\d\d\d)_/) {
				$layer = sprintf("%s.%03d_negative%s", $1, 999-$2, substr($layer,7));
			}
		}
		#$layer = 'depression_line' unless defined($ok_contours{$layer});
		
		if ($POLY) {
			printf( DXF "0\nPOLYLINE\n66\n1\n8\n%s\n10\n0\n20\n0\n30\n%1.2f\n38\n%1.2f\n", 
				$layer, $height, $height);
			my $VERTEX = "0\nVERTEX\n8\n$layer\n";
			foreach my $rec (@points) {
				my ( $x, $y, $h ) = split( /\,/, $rec );
					
				printf( DXF "%s10\n%1.2f\n20\n%1.2f\n", $VERTEX, $x*$XINV, $y*$YINV);
			}
		}
		else {
			my $LOOP = key($points[0]) eq key($points[-1]) ? "70\n1\n" : '';
			printf( DXF "0\nLWPOLYLINE\n8\n%s\n38\n%1.2f\n%s90\n%d\n", $layer, $height, $LOOP, scalar(@points));
			foreach my $rec (@points) {
				my ( $x, $y, $h ) = split( /,/, $rec );
					
				printf( DXF "10\n%1.2f\n20\n%1.2f\n", $x*$XINV, $y*$YINV);
			}
		}
		printf( DXF "0\nSEQEND\n");
	}
	if (defined($point_depressions)) {
		foreach (@$point_depressions) {
			my ( $x, $y, $dir ) = split(/,/);
			printf( DXF "0\nPOINT\n8\n%s\n10\n%1.2f\n20\n%1.2f\n", '115_depression', 
				$x*$XINV, $y*$YINV);
		}
	}
	if (defined($dot_knolls)) {
		foreach (@$dot_knolls) {
			my ( $x, $y, $h ) = split(/,/);
			printf( DXF "0\nPOINT\n8\n%s\n38\n%1.2f\n10\n%1.2f\n20\n%1.2f\n", 
				'112_dot_knoll', 
				$h, 
				$x*$XINV, $y*$YINV);
		}
	}
	if (defined($slope_lines)) {
		foreach (@$slope_lines) {
			my ( $x, $y, $dir ) = split(/,/);
			printf( DXF "0\nPOINT\n8\n%s\n10\n%1.2f\n20\n%1.2f\n50\n%1.2f\n", '104_slope_line', 
				$x*$XINV, $y*$YINV, $dir);
		}
	}
	printf( DXF "0\nENDSEC\n0\nEOF\n");
	close(DXF);
	#foreach (sort {$a <=> $b} keys %height2layer) {
	#	printf(STDERR "%1.5f : %s\n", $_, $height2layer{$_});
	#}
}

sub write_txt
{
	my ($txt, $xmin, $xmax, $ymin, $ymax, $dep, $knolls, $slopes, $contours) = @_;
	
	open(TXT, ">", $txt) || die("Could not write to $txt\n");
	
	foreach (@$dep) {
		my ($x, $y, $dir) = split(/,/);
		printf(TXT "%1.2f,%1.2f,115_dep\n#\n", $x*$XINV, $y*$YINV);
	}
	foreach (@$knolls) {
		my ($x, $y, $h) = split(/,/);
		printf(TXT "%1.2f,%1.2f,112_knoll_%1.2f\n#\n", $x*$XINV, $y*$YINV, $h*$HINV);
	}
	foreach (@$slopes) {
		my ($x, $y, $dir) = split(/,/);
		printf(TXT "%1.2f,%1.2f,104_slope_%d\n#\n", $x*$XINV, $y*$YINV, $dir);
	}
	foreach (@$contours) {
		my @points = split(/\n/);
		foreach (@points) {
			my ($x, $y, $h) = split(/,/);
			printf(TXT "%1.2f,%1.2f,%1.2f\n", $x*$XINV, $y*$YINV, $h*$HINV);
		}
		printf(TXT "#\n");
	}
	close(TXT);
}

sub intercept
{
	my ($y0, $xstep, $slope) = @_;
	my $y1 = sprintf("%1.0f", $y0 + $xstep * $slope);
	return $y1;
}

sub clip_contours
{
	my ($xmin, $xmax, $ymin, $ymax, @contours) = @_;
	my @c = ();
	$xmin *= $XSCALE;
	$xmax *= $XSCALE; 
	$ymin *= $YSCALE;
	$ymax *= $YSCALE; 
	
	
	foreach (@contours) {
		my $c = $_;
		my @points = split(/\n/);

		my ($inside, $prev_inside);
		my @p = ();
		my ($xprev, $yprev);

		my $outside_only = 1;  # Skip this contour if all points are outside!
		my $clips = 0;
		foreach (@points) {
			my ($x, $y, $h) = split(/,/);
			next unless defined($h);
			
			$inside = 0;
			if ($x >= $xmin && $x <= $xmax && $y >= $ymin && $y <= $ymax) {
				$inside = 1;
				$outside_only = 0;
			}
			$prev_inside = $inside unless defined($prev_inside);
			
			if ($inside) { # This point is inside!
				if ($prev_inside) { # and the entire line segment is inside
					push(@p, $_);  # so save this point
				}
				else { # From outside to inside transition, starting a new segment at the boundary line:
					my ($xb, $yb);
					my ($dx, $dy) = ($x - $xprev, $y-$yprev);
					# Check vertical xmin/xmax boundaries:
					if ($xprev < $xmin) {
						#$yb = sprintf("%1.0f", ($yprev + ($xmin-$xprev)*$dy/$dx));
						$yb = intercept($yprev, $xmin-$xprev, $dy/$dx);
						$xb = $xmin;
						$xprev = $xb;
						$yprev = $yb;
					}
					elsif ($xprev > $xmax) {
						#$yb = sprintf("%1.0f", ($yprev + ($xmax-$xprev)*$dy/$dx));
						$yb = $yb = intercept($yprev, $xmax-$xprev, $dy/$dx);
						$xb = $xmax;
						$xprev = $xb;
						$yprev = $yb;
					}
					# Ditto for horizontal ymin/ymax limits:
					if ($yprev < $ymin) {
						#$xb = sprintf("%1.0f", ($xprev + ($ymin-$yprev)*$dx/$dy));
						$xb = intercept($xprev, $ymin-$yprev, $dx/$dy);
						$yb = $ymin;
					}
					elsif ($yprev > $ymax) {
						#$xb = sprintf("%1.0f", ($xprev + ($ymax-$yprev)*$dx/$dy));
						$xb = intercept($xprev, $ymax-$yprev, $dx/$dy);
						$yb = $ymax;
					}
					push(@p, join(',', $xb, $yb, $h));	# Boundary point
					push(@p, join(',', $x, $y, $h));	# current point
					#printf(STDERR "%d", ++$clips);
				}
			}
			elsif ($prev_inside) { # Inside -> outside
				my ($xb, $yb);
				my ($xs, $ys) = ($x, $y);
				my ($dx, $dy) = ($x - $xprev, $y-$yprev);
				# Vertical xmin/xmax limits:
				if ($x < $xmin) {
					$xb = $xmin;
					#$yb = sprintf("%1.0f", ($yprev + ($xmin-$xprev)*$dy/$dx));
					$yb = intercept($yprev, $xmin-$xprev, $dy/$dx);
					$x = $xb;
					$y = $yb;
				}
				elsif ($x > $xmax) {
					$xb = $xmax;
					#$yb = sprintf("%1.0f", ($yprev + ($xmax-$xprev)*$dy/$dx));
					$yb = intercept($yprev, $xmax-$xprev, $dy/$dx);
					$x = $xb;
					$y = $yb;
				}
				if ($y < $ymin) {
					$yb = $ymin;
					#$xb = sprintf("%1.0f", ($xprev + ($ymin-$yprev)*$dx/$dy));
					$xb = intercept($xprev, $ymin-$yprev, $dx/$dy);
				}
				elsif ($y > $ymax) {
					$yb = $ymax;
					#$xb = sprintf("%1.0f", ($xprev + ($ymax-$yprev)*$dx/$dy));
					$xb = intercept($xprev, $ymax-$yprev, $dx/$dy);
				}
				push(@p, join(',', $xb, $yb, $h));	# Boundary point
				push(@c, join("\n", @p));
				@p = ();
				#printf(STDERR "%d", ++$clips);
				($x, $y) = ($xs, $ys);
			}
			$xprev = $x; $yprev = $y; $prev_inside = $inside;
		}
		if (scalar(@p) > 1) {
			push(@c, join("\n", @p));
		}
		elsif ($outside_only) { # outside only, so skip!
			#printf(STDERR "-");
		}
	}
	#printf(STDERR "\n%d contours after clipping.\n", scalar(@c));
	return @c;
}

sub key
{
	my ($p) = @_;
	my ($x, $y, $h) = split(/,/, $p);
	$p = sprintf("%1.0f,%1.0f,%s", $x*$XINV, $y*$YINV, $h);
	return $p;
}

sub akeys
{
	my ($dist, $p) = @_;
	
	my ($xs, $ys, $h) = split(/,/, $p);
	my @xs = ();
	my @ys = ();
	# Convert point coordinates to meter
	$xs *= $XINV;
	$ys *= $YINV;

	# Accept matches within the nearest meter:
	@xs = (int($xs), int($xs)+1);
	@ys = (int($ys), int($ys)+1);

	my @p = ();
	foreach (@ys) {
		my $y = $_;
		foreach (@xs) {
			my $x = $_;
			$p = sprintf("%1.0f,%1.0f,%s", $x, $y, $h);
			push(@p, $p);
		}
	}
	return @p;
}

sub merge_contours
{
	my @contours = @_;
	
	printf(STDERR "\n%s%d contours before merging\n", $CID, scalar(@contours));
	my %heads = ();
	my %tails = ();
	my $loops = 0;
	my @loops = ();
	
	for (my $c = 0; $c < scalar(@contours); $c++) {
		my @points = split(/\n/, $contours[$c]);
		# Is this a loop? If so, skip!
		if (key($points[0]) eq key($points[scalar(@points)-1])) {
			$loops++; 
			push(@loops, $contours[$c]);
			undef $contours[$c];
			next;
		}
		my ($head, $tail, $found);
		
		do {
			$head = key($points[0]);
			$tail = key($points[-1]);
			$found = 0;
			
			# Do we need to reverse this segment?
			if (defined($heads{$head}) || defined($tails{$tail})) {
				@points = reverse @points;
				my $t = $tail;
				$tail = $head;
				$head = $t;
				# Will match in one of the two next tests!
			}
			if (defined($heads{$tail})) {
				# This line is in front of an existing one:
				my $h = $heads{$tail};
				my @p = split(/\n/, $contours[$h]);
				# Use dist instead! die("No match!\n") unless ($tail eq key($p[0]));
				
				# Remove all traces of previous segment:
				undef $contours[$h];
				foreach (akeys(1,$p[0])) {
					$heads{$_} = undef;
				}
				foreach (akeys(1,$p[-1])) {
					$tails{$_} = undef;
				}
				
				shift @p; # Get rid of the extra/identical boundary point!
				push(@points, @p);
				#$tail = key($points[-1]);
				
				$found++;
			}
			
			if (defined($tails{$head})) {
				# This line follows an existing one:
				my $h = $tails{$head};
				my @p = split(/\n/, $contours[$h]);
				# dist instead! die("No match!\n") unless ($head eq key($p[-1]));
				
				undef $contours[$h];
				foreach (akeys(1,$p[0])) {
					$heads{$_} = undef;
				}
				foreach (akeys(1,$p[-1])) {
					$tails{$_} = undef;
				}
				
				shift @points;
				@points = (@p, @points);
				#$head = $points[0];
				
				$found++;
			} 
			#printf(STDERR "%d", $found) if ($found);
			printf(STDERR "%s\r", ('-','\\','|','/')[$c & 3]) if ($tid);
		} while ($found);
		
		my $cont = join("\n", @points);
		if (key($points[0]) eq key($points[-1])) {
			$loops++; 
			push(@loops, $cont);
			undef $contours[$c];
			#$heads{$head} = $tails{$tail} = undef;
		}
		else {
			$contours[$c] = $cont;
			foreach (akeys(1,$points[0])) {
				$heads{$_} = $c;
			}
			foreach (akeys(1,$points[-1])) {
				$tails{$_} = $c;
			}
		}
	}
	my @c = ();
	foreach (keys %heads) {
		if (defined($heads{$_}) && defined($contours[$heads{$_}])) {
			push(@c, $contours[$heads{$_}]);
			undef $contours[$heads{$_}];
		}
	}
	printf(STDERR "%s%d contours after merging (%d loops)\n", $CID, scalar(@c)+$loops, $loops);
	return (@loops, @c); # Return all the loops first!
}

sub myglob
{
	my @match = ();
	foreach (@_) {
		my $dir = '.';
		if (/^(.+)[\\\/]([^\\\/]+)$/) { # Starts with a directory!
			$dir = $1;
			$_ = $2;
		}
		s/\./\t/g;		# Save away any explicit '.' symbols
		s/\*/\.*/g;		# Convert * to .*
		s/\?/\./g;		# Convert ? to .
		s/\t/\\\./g;	# Restore explicit '.' as escaped '\.'
		
		my $pattern = $_;
		
		opendir(D, $dir);
		my @files = readdir(D);
		closedir(D);
		
		$dir = ($dir eq '.') ? '' : $dir .'/';
		
		foreach (@files) {
			next unless (-f $dir.$_);
			
			next unless (/^$pattern$/);
			push(@match, $dir . $_);
		}
	}
	return @match;
}

sub gcd
{
	my @c = @_;
	
	return undef unless (scalar(@c));
	
	@c = sort {$a <=> $b} @c;
	
	my $gcd = $c[0];
	my $scale;
	do {
		my $min_fraction = 1;
		foreach (@c) {
			my $fraction = $_ / $gcd;
			$fraction -= int($fraction);
			if ($fraction > 0.001 && $fraction < $min_fraction) {
				$min_fraction = $fraction;
			}
		}
		$scale = sprintf("%1.0f", 1/$min_fraction);
		$gcd /= $scale;
	} while ($scale > 1);
	printf(STDERR "Contours at %s, GCD = %1.2f\n", join(",", @c), $gcd);
	return $gcd;
}

my ($ncols, $nrows, $xll, $yll, $cell, $NODATA, @alt);

sub load_asc
{
	my ($fname) = @_;
	open(S, $fname) || die("Could not open $fname\n");
	my %tags;
	for (my $hdr = 0; $hdr < 6; $hdr++) {
		$_ = <S>; 
		chomp; 
		my ($tag, $val) = split;
		$tags{$tag} = $val;
	}
	$ncols = $tags{ncols};
	$nrows = $tags{nrows};
	$xll = $tags{xllcorner};
	$yll = $tags{yllcorner};
	$cell = $tags{cellsize};
	$NODATA = $tags{NODATA_value};

	@alt = ();
	while (<S>) {
		chomp;
		push(@alt, $_);
	}
	close(S);
	return ($ncols, $nrows, $xll, $yll, $cell, $NODATA, @alt);
}

sub altitude
{
	my ($x, $y) = @_;
	my $xoff = int(($x - $xll) / $cell + 0.5);
	
	if ($xoff < 0 || $xoff >= $ncols) {
		#warn("x = $x ($xoff) is out of range") ;
		return $NODATA;
	}
	my $line = int(($y - $yll) / $cell + 0.5);
	$line = $nrows - 1 - $line;
	if ($line < 0 || $line >= $nrows) {
		#warn("y = $y ($line) is out of range");
		return $NODATA;
	}
	my $h = (split(' ',$alt[$line]))[$xoff];
	return $h;
}

sub find_depressions
{
	my ($ASC, @contours) = @_;
	load_asc($ASC);
	my (@c, @slope_lines, @point_depressions) = ();
	foreach(@contours) {
		my $cont = $_;
		my @points = split(/\n/);
		my $found = 0;
		if (key($points[0]) ne key($points[-1]) ||
			(scalar(@points) > 100)) {
			push(@c, $cont);
			next;
		}
				
		my ($xmin, $xmax, $ymin, $ymax) = (1e38,-1e38,1e38,-1e38);
		my ($xminy, $xmaxy, $yminx, $ymaxx);
		my ($x, $y, $h, $px, $py);
		my $len = 0;
		foreach (@points) {
			($x, $y, $h) = split(/,/);
			($x, $y, $h) = ($x*$XINV, $y*$YINV, $h*$HINV);

			if ($x < $xmin) {
				$xmin = $x ;
				$xminy = $y;
			}
			if ($x > $xmax) {
				$xmax = $x;
				$xmaxy = $y;
			}
			if ($y < $ymin) {
				$ymin = $y;
				$yminx = $x;
			}
			if ($y > $ymax) {
				$ymax = $y;
				$ymaxx = $x;
			}
			
			if (defined($px)) {
				$len += sqrt(($x-$px)*($x-$px) + ($y-$py)*($y-$py));
			}
			($px, $py) = ($x, $y);
		}
		if ($len > 100) { # Skip very long contours, don't need tag lines
			push(@c, $cont);
			next;
		}
			
		# Do 3+ of the 4 extremal points lean inwards?
		my $vote = 0;
		$vote += altitude($xmin, $xminy) > altitude($xmin+1,$xminy);
		$vote += altitude($xmax, $xmaxy) > altitude($xmax-1,$xmaxy);
		$vote += altitude($yminx, $ymin) > altitude($yminx,$ymin+1);
		$vote += altitude($ymaxx, $ymax) > altitude($ymaxx,$ymax-1);
		if ($vote < 3) {
			push(@c, $cont);
			next;
		}
		# Is the depression too small for a contour?
		my $dy = $ymax-$ymin;
		my $dx = $xmax-$xmin;
		if ($dy < 12 && $dx < 12) { # 7 m tags in 1:10000!
			push(@point_depressions, sprintf("%1.2f,%1.2f,%1.2f", ($xmin+$xmax)*0.5, ($ymin+$ymax)*0.5, $h));
			# Remove it from the contour list!
		}
		else {
			push(@c, $cont);
			# (0, 90, 180, -90) is (up, left, down, right)
			if ($dx > $dy) {
				push(@slope_lines, sprintf("%1.2f,%1.2f,%1.2f", $xmin, $xminy, -90));
				if ($dx > 21) {
					push(@slope_lines, sprintf("%1.2f,%1.2f,%1.2f", $xmax, $xmaxy, 90));
				}
			}
			else {
				push(@slope_lines, sprintf("%1.2f,%1.2f,%1.2f", $yminx, $ymin, 0));
				if ($dy > 21) {
					push(@slope_lines, sprintf("%1.2f,%1.2f,%1.2f", $ymaxx, $ymax, 180));
				}
			}
		}
	}
	return (\@point_depressions, \@slope_lines, @c);
}
