#!perl -w

use strict;
use Cwd;
use GD;

my $laz;
my $output_file;
my $FORCE = 0;
my $FACIT_FILE = "vegetasjon_facit.txt";
my $FACIT_SCALE;
my $CORES = 1; # If > 1, run separate threads for the tile processing
my $KEEP_RAW = 0; # Enable to write raw*.gif files

my $SAMPLE = 5;
my $SAMPLE_START = -2;
my $SAMPLE_SLUTT = 2;

while ($_ = shift) {
	if (/^-force$/i) {
		$FORCE = 1;
	}
	elsif (/^-facit$/i) {
		$FACIT_FILE = shift;
	}
	elsif (/^-raw$/i) {
		$KEEP_RAW = 1;
	}
	elsif (/^-cores$/i) {
		$CORES = shift;
		if (! ($CORES =~ /^\d+$/)) {
			$CORES = 1;
		}
		$FORCE = 1 if ($CORES > 1);
	}
	elsif (/^-sample$/i) {
		$SAMPLE = int(shift);
		$SAMPLE++ if ($SAMPLE & 1 == 0); # Make it odd!
		$SAMPLE_SLUTT = $SAMPLE >> 1;
		$SAMPLE_START = -$SAMPLE_SLUTT;
	}
	elsif (!defined($laz)) {
		$laz = $_;
	}
	elsif (!defined($output_file)) {
		$output_file = $_;
	}
}

# Pick up current directory name, use as default output base:
if (! defined($output_file)) {
	$output_file = getcwd;
	if ($output_file =~ /^(.+)[\\\/]([^\\\/]+)$/) {
		$output_file = $2;
	}
	$output_file = $output_file . '_vegetation.gif'
}
$output_file = $output_file . '.gif' unless ($output_file =~ /\.gif$/);

my @files = myglob($laz);

if (scalar(@files) == 1 && ! ($files[0] =~ /\.la[sz]$/i)) { # single non-LA[SZ] file, use as List of Files:
	open(LOF, $files[0]);
	@files = ();
	while (<LOF>) {
		chomp;
		push(@files, $_);
	}
	close(LOF);
}

die("Usage: gen-laz-veg.pl lazfile|file_with_list_of_laz_files output_file\n") unless (scalar(@files) && defined($output_file));

my ($XMIN, $XMAX, $YMIN, $YMAX, $STEP) = find_min_max(@files);

my @tiles = ();
foreach (sort @files) {
	if (/tile_([\-]?\d+)_([\-]?\d+)\D/) {
		my ($xmin, $xmax, $ymin, $ymax) = ($1, $1+$STEP-1, $2,$2+$STEP-1);
		push(@tiles, join(',', $_, $xmin, $xmax, $ymin, $ymax));
	}
}

die("No laz/las files found!\n") unless scalar(@tiles);

# Points of type 2 (Ground), 3/4/5 (Low/Med/High vegetation) are accumulated, ignore all others
my @CLASS_INCREMENT = (0,0,1,4096,4096*4096,4096*4096*4096, (0) x 10);

my $SCALE = 2;
my $INVSCALE = 1.0/$SCALE;

my @SCALE_FACTOR = (1,1,1,1);

my %facit = ();
my %class2mask = ();
my $usefacit = 0;
if (-f $FACIT_FILE) {
	$usefacit = 1;
	open(F, $FACIT_FILE);
	while (<F>) {
		chomp;
		my ($e, $n, @f) = split(/,/);
		if ($f[0] eq 'M') {
			$FACIT_SCALE = $_;
			for (my $i = 0; $i < 4; $i++) {
				$SCALE_FACTOR[$i] = 100 / $f[$i+1];
			}
			next;
		}
		$facit{scale($e).';'.scale($n)} = join(',',@f);
		if (scalar(@f) < 5) {
			$usefacit = -1;  # Needs updates!
			$CORES = 1;
		}
	}
	close(F);
	printf(STDERR "%s $FACIT_FILE\n", $usefacit > 0 ? 'Using' : 'Updating');
}

sub scale
{
  my ($x) = @_;
  $x = int($x * $INVSCALE);
  return $x;
}

my ($X0, $Y0, $X1, $Y1) = (scale($XMIN), scale($YMIN), scale($XMAX), scale($YMAX));
printf("Total area: (%d,%d)-(%d,%d)\n", $XMIN, $YMIN, $XMAX, $YMAX);

# Generate sampling matrix:
my @vekt = ();
for (my $dy = $SAMPLE_START; $dy <= $SAMPLE_SLUTT; $dy++) {
	for (my $dx = $SAMPLE_START; $dx <= $SAMPLE_SLUTT; $dx++) {
		my $vekt = $SAMPLE*$SAMPLE*0.25 - ($dx*$dx + $dy*$dy); # Quadratic sample weight falloff
		#my $vekt = $SAMPLE - sqrt($dx*$dx + $dy*$dy)*2;
		$vekt = 0 if ($vekt < 0.0);
		$vekt[$dy-$SAMPLE_START][$dx-$SAMPLE_START] = $vekt;
		printf(STDERR "%5.2f", $vekt);
	}
	printf(STDERR "\n");
}
printf(STDERR "\n");

# Majority sampling matrix:
my $SAMPLE_DIST = 2;
my @vekt_major = ();
for (my $dy = -$SAMPLE_DIST; $dy <= $SAMPLE_DIST; $dy++) {
	for (my $dx = -$SAMPLE_DIST; $dx <= $SAMPLE_DIST; $dx++) {
		my $vekt = $SAMPLE_DIST - sqrt($dx*$dx + $dy*$dy)*0.5;
		$vekt = 0 if ($vekt < 0.0);
		$vekt_major[$dy+$SAMPLE_DIST][$dx+$SAMPLE_DIST] = $vekt;
	}
}

sub setPalette
{
	my ($raw) = @_;
	$raw->colorAllocate(255,255,255);	# 0 white
	$raw->colorAllocate(0,0,0);			# 1 black
	$raw->colorAllocate(162,149,179);	# 2 green_stripe
	$raw->colorAllocate(255,149,179);	# 3 yellow_with_green_stripes
	$raw->colorAllocate(255,219,166);  	# 4 yellow color="#ffdba6" (255,242,0) color="#ffff00" (jorde: color="#ffc000")
	$raw->colorAllocate(160,255,160);	# 5 light green color="#a0ffa0"  color="#7fff7f"
	$raw->colorAllocate(127,255,127);	# 6 green color="#1eff1e" color="#00ff00"
	$raw->colorAllocate(0,255,0);		# 7 dark green color="#008000"
}

sub setPaletteFinal
{
	my ($raw) = @_;
	$raw->colorAllocate(255,255,255);	# 0 white
	$raw->colorAllocate(0,0,0);			# 1 black
	$raw->colorAllocate(255,255,255);	# 2 white for green_stripe
	$raw->colorAllocate(255,219,166);	# 3 yellow for green_stripes
	$raw->colorAllocate(255,219,166);  	# 4 yellow color="#ffdba6" (255,242,0) color="#ffff00" (jorde: color="#ffc000")
	$raw->colorAllocate(160,255,160);	# 5 light green color="#a0ffa0"  color="#7fff7f"
	$raw->colorAllocate(127,255,127);	# 6 green color="#1eff1e" color="#00ff00"
	$raw->colorAllocate(0,255,0);		# 7 dark green color="#008000"
}

my ($tid, @tids) = (1);

my $CID = '';
if ($CORES > 1) { # Start sub-processes!
	for (my $core = 0; $core < $CORES; $core++) {
		my @sub_tiles = ();
		for (my $f = $core; $f < scalar(@tiles); $f += $CORES) {
			push(@sub_tiles, $tiles[$f]);
		}
		$tid = fork;
		if ($tid == 0) { # This is the forked sub-thread!
			$CID = "$core: ";
			my $s = 0;
			foreach (@sub_tiles) {
				my ($laz, $xmin, $xmax, $ymin, $ymax) = split(/,/);
			
				$s++;
				printf(STDERR "%stile %d/%d %s\n", $CID, $s, scalar(@sub_tiles), $laz);
			
				my $bm = generate_multi($laz, $xmin, $xmax, $ymin, $ymax);
			}
			die("\n$CID Sub-thread done!\n\n");
		}
		printf(STDERR "PID $tid forked\n");
		push(@tids, $tid);
	}
	foreach (@tids) {
		my $kid = waitpid($_, 0);
		#sleep(1);
	}
	printf(STDERR "Finished waiting, collecting raw tiles\n");
	$FORCE = 0;
}

my ($final, $finalraw);

# Collect all the final gif images here, starting at X0, Y0!
printf(STDERR "Creating final target image of (%dx%d) pixels, starting at (%d,%d)\n", $XMAX-$XMIN+1, $YMAX-$YMIN+1, $XMIN, $YMIN);
$final = GD::Image->new($XMAX-$XMIN+1, $YMAX-$YMIN+1);
setPaletteFinal($final);

if ($KEEP_RAW) {
	$finalraw = GD::Image->new(($XMAX-$XMIN+1)/2, ($YMAX-$YMIN+1)/2);
	setPalette($finalraw);
}

my $s = 0;
foreach (@tiles) {
	my ($laz, $xmin, $xmax, $ymin, $ymax) = split(/,/);
	
	printf(STDERR "%s%d/%d %s\n", $CID, ++$s, scalar(@tiles), $laz);
	
	my $gifname = substr($laz,0,length($laz)-3).'gif';
	my $bm;
	if (-f $gifname && (($FORCE == 0) || ($tid && scalar(@tids)))) { # We have used fork previously!
		$bm = GD::Image->newFromGif($gifname);
		#print $bm;
	}
	else {
		$bm = generate_multi($laz, $xmin, $xmax, $ymin, $ymax);
		next if ($usefacit == -1);
	}
	my ($bx, $by) = $bm->getBounds;
	if ($laz =~ /tile_([\-]?\d+)_([\-]?\d+)\D/) {
		die("Tile name ($laz) does not match location $xmin, $ymin!\n") unless ($xmin == $1 && $ymin == $2);
	}
	my ($xoffs, $yoffs) = ($xmin-$XMIN, $YMAX - $ymax);
	printf(STDERR "%scopying to $xoffs, $yoffs\n", $CID);
	$final->copy($bm, $xoffs, $yoffs, 0, 0, $bx, $by);
	if ($KEEP_RAW) {
		$gifname =~ s/tile_/raw_/;
		$bm = GD::Image->newFromGif($gifname);
		($bx, $by) = $bm->getBounds;

		$finalraw->copy($bm, $xoffs/2, $yoffs/2, 5, 5, 125, 125);
	}
}

if ($usefacit == -1) { # Updating...
	open(F, ">$FACIT_FILE") || die("Could not update $FACIT_FILE");
	if (defined($FACIT_SCALE)) {
		printf(F "%s\n", $FACIT_SCALE);
	}
	foreach (sort keys %facit) {
		my @f = $facit{$_};
		my ($e, $n) = split(/;/, $_);
		printf(F "%d,%d,%s\n", $e*$SCALE, $n*$SCALE, join(',', @f));
	}
	close(F);
	die("$FACIT_FILE updated!\n");
}

my $gifname = $output_file;

open G,">",$gifname or die "$!";
binmode G;
print G $final->gif;
close(G);

my $world = substr($gifname,0,length($gifname)-3).'gfw';
open(W, ">", $world) or die "$!";
$SCALE = 1;
$X0 *= $SCALE;
$Y1 = ($Y1+1)*$SCALE-1;
foreach (1, 0, 0, -1, 
         ($XMIN+0.5), 
		 ($YMAX-0.5)
		) {
  printf(W "%1.1f\n", $_);
}
close(W);

if ($KEEP_RAW) {
	$gifname = 'raw_'.$output_file;

	open G,">",$gifname or die "$!";
	binmode G;
	print G $finalraw->gif;
	close(G);

	my $world = substr($gifname,0,length($gifname)-3).'gfw';
	open(W, ">", $world) or die "$!";
	foreach (2, 0, 0, -2, 
			 ($XMIN+1), 
			 ($YMAX-1)
			) {
	  printf(W "%1.1f\n", $_);
	}
	close(W);
}

printf(STDERR "Vegetation generation done!\n");

sub generate_multi # Both raw and final GIFs
{
	my ($laz, $mine, $maxe, $minn, $maxn) = @_;

	my $BUFFER = 10;
	my $SCALED_BUFFER = scale($BUFFER);
	my ($mines, $maxes, $minns, $maxns) = (scale($mine-$BUFFER), scale($maxe+$BUFFER), scale($minn-$BUFFER), scale($maxn+$BUFFER));
	
	if ($usefacit == -1) {
		my $facit_inside = 0;
		foreach (keys %facit) {
			my ($es, $ns) = split(/;/);
			if ($es >= $mines && $es < $maxes && $ns >= $minns && $ns < $maxns) {
				$facit_inside = 1;
				last;
			}
		}
		return unless ($facit_inside);
	}
	
	my $dir = '';
	if ($laz =~ /^(.+[\\\/])([^\\\/]+)$/) { # Look for last dir separator!
		$dir = $1;
		$laz = $2;
	}
	my $gifname = $dir.'raw'.substr($laz,4,length($laz)-8).'.gif';
	$laz = $dir.$laz;

	my ($bm, $percent, $count, $percent_per_line);

	printf(STDERR "$CID%s: dimensions: %dx%d\n", $gifname, $maxes-$mines+1, $maxns-$minns+1);
	
	$bm = GD::Image->new($maxes-$mines+1, $maxns-$minns+1);  # Includes room for a 10 cell border all the way around
	setPalette($bm);

	my @c = ();	# Current counts
	my ($x0, $y0) = (scale($mine)-25, scale($minn)-25);	# Include room for 25-cell guard!

	my $ll = $laz;
	$ll  =~ s/\//\\/g;
	open(I, "-|", "las2txt -i $ll -parse xyzc -sep space -stdout") || die ("Could not open pipe from las2txt!\n");
	printf(STDERR "%sPiping point classifications directly from las2txt\n", $CID);
	while (<I>) {
		chomp;
		my ($e, $n, $h, $class) = split;
		next unless (defined($class) && ($class ne ''));
	  
		my ($es, $ns) = (scale($e), scale($n));
	  
		my $inc = $CLASS_INCREMENT[$class];
		$c[$ns-$y0][$es-$x0] += $inc;
	}
	close(I);
	#printf(STDERR "Done!\n");

#	setPalette($bm);
	
	my $white = 0;
	
	$bm->rectangle(0,0,$maxes-$mines, $maxns-$minns, $white);
	$bm->fill(1,1,$white);

	#my @colors = ($yellow, $yellow_green_stripe, $light_green, $green, $white, $green_stripe, $green, $dark_green);
	my @colors = (4, 3, 5, 6, 0, 2, 6, 7);
	my @greenmask = (0,0,0,0,0,1,1,1); # To aggregate the various green colors...
	%class2mask = ('Y' => 0, 'YS' => 1, 'LG' => 2, 'G' => 3, 'W' => 4, 'WS' => 5, 'DG' => 7);
	
	#printf(STDERR "%sWriting raw vegetation file\n", $CID);
	for (my $es = $mines; $es <= $maxes; $es++) {	
		for (my $ns = $minns; $ns <= $maxns; $ns++) {
			my $hist = 0;
			my @hist = (0,0,0,0);
			for (my $e = $SAMPLE_START; $e <= $SAMPLE_SLUTT; $e++) {
				for (my $n = $SAMPLE_START; $n <= $SAMPLE_SLUTT; $n++) {
					my $k = sprintf("%d;%d", $es+$e, $ns+$n);
					$hist = $c[$ns+$n-$y0][$es+$e-$x0];
					$hist += 0; # In case of undef!
					my @h = ($hist & 0xfff, ($hist >> 12) & 0xfff, ($hist >> 24) & 0xfff, ($hist >> 36) & 0xff);
					my $vekt = $vekt[$n-$SAMPLE_START][$e-$SAMPLE_START];
					printf(STDERR "vekt[%d][%d] = $vekt\n", $n-$SAMPLE_START, $e-$SAMPLE_START) unless (defined($vekt));
					for (my $i = 0; $i < 4; $i++) {
						$hist[$i] += $h[$i] * $vekt;
					}
				}
			}
			#if (defined($facit{"$es;$ns"})) {
			#	printf("%d;%d;%s\n", $es*$SCALE, $ns*$SCALE, join(" ", @hist));
			#}
			my $cnt = 0;
			my $mask = 4;  # White forest is the default
			foreach (@hist) { $cnt += $_; }
			if ($cnt) {
				my $scale = 100 / $cnt;
				for (my $i = 0; $i < 4; $i++) {
					$hist[$i] *= $scale;
				}
				if ($usefacit == -1 && defined($facit{"$es;$ns"})) {
					$facit{"$es;$ns"} = join(',', (split(/,/,$facit{"$es;$ns"}))[0], map {sprintf("%1.1f", $_)} @hist);
					printf("%d,%d,%s\n", $es*$SCALE, $ns*$SCALE, $facit{"$es;$ns"});
				}
				if ($hist[0] > 98) { 
					$mask = 0;	# Hardcoded Yellow open land!
				}
				else {
					my ($mindist, $minclass) = (1e38);
					if ($usefacit > 0) {
						foreach (keys %facit) {
							my ($class, @f) = split(',', $facit{$_});
							my $dist = 0;
							for (my $i = 0; $i < 4; $i++) {
								my $d = ($f[$i]-$hist[$i])*$SCALE_FACTOR[$i];
								$dist += $d*$d;
							}
							if ($dist < $mindist) {
								$minclass = $class;
								$mindist = $dist;
							}
						}
					}
					if (defined($minclass)) {
						die("Unknown vegetation class $minclass!") unless defined($class2mask{$minclass});
						$mask = $class2mask{$minclass};
					}
					else {
						$mask = 0;
						if ($hist[3] > 10) { # High vegetation
							$mask = 4;
							if ($hist[1] > 10) { $mask |= 1;}
							if ($hist[2] > 10) { $mask |= 2;}
						}
						else {
							if ($hist[1] > 25) { $mask |= 1;}
							if ($hist[2] > 25) { $mask |= 6;}
						}
					}
				}
			}
			$bm->setPixel($es-$mines, $maxns-$ns, $colors[$mask]);
		}
		printf(STDERR "%2d%%\r", ($es-$mines)*100/($maxes-$mines)) if ($tid);
	}

	if ($KEEP_RAW) {
		open( G, ">", $gifname) or die "$!";
		binmode G;
		print G $bm->gif;
		close(G);

		my $world = substr($gifname,0,length($gifname)-3).'gfw';
		open(W, ">", $world) or die "$!";
		foreach ($SCALE, 0, 0, -$SCALE, 
				 ($mines+0.5) * $SCALE, 
				 ($maxns+0.5) * $SCALE
				) {
		  printf(W "%1.4f\n", $_);
		}
		close(W);
	}
	# Create filtered and cropped final image:
	
	my $vote = GD::Image->new($maxe-$mine+1, $maxn-$minn+1);
	setPaletteFinal($vote);

	($mines, $maxes, $minns, $maxns) = (scale($mine), scale($maxe), scale($minn), scale($maxn));
	# Pick (center-weighted) majority pixel value for each 7x7 (14x14m) area:
	$percent = 0;
	$count = 0;
	$percent_per_line = 100 / ($maxns-$minns+1);

	$gifname = substr($laz,0,length($laz)-4).'.gif';
	printf(STDERR "%sWriting majority vegetation file $gifname\n", $CID);
	
	for (my $n = 0; $n <= $maxns-$minns; $n++) {
		my $ns = $n-$SAMPLE_DIST;
		my $ne = $n+$SAMPLE_DIST;
		for (my $e = 0; $e <= $maxes-$mines; $e++) {
			my %counts = ();
			my $es = $e-$SAMPLE_DIST;
			my $ee = $e+$SAMPLE_DIST;
			for (my $sn = $ns; $sn <= $ne; $sn++) {
				for (my $se = $es; $se <= $ee; $se++) {
					my $idx = $bm->getPixel($se+$SCALED_BUFFER, $sn+$SCALED_BUFFER);
					$counts{$idx}+= $vekt_major[$sn-$n+$SAMPLE_DIST][$se-$e+$SAMPLE_DIST];
				}
			}
			my ($greensum, $idx, $max) = (0,0,0);
			foreach (keys %counts) {
				$greensum += $counts{$_} * $greenmask[$_];
			}
			$counts{8} = $greensum;
			my @counts = reverse sort {$counts{$a} <=> $counts{$b}} keys %counts;
			$idx = shift @counts;
			if ($idx > 7) { # Green* wins, select best type:
				do {
					$idx = shift @counts; 
				} while ($greenmask[$idx] == 0);
			}
			die("Out of range: idx = $idx for pos = $e,$n!\n") if ($idx < 0 || $idx > 7);
			#$total_counts{$idx}++;
			if ($SCALE == 2) {
				my $idx2 = $idx;
				if ($e & 1 && ($idx == 2 || $idx == 3)) { # 25% Green vertical stripes
					$idx = 6; # Green in 0010 pattern: All even pixels have original colour!
				}
				$vote->setPixel(2*$e + 0, 2*$n + 0, $idx);
				$vote->setPixel(2*$e + 0, 2*$n + 1, $idx);
				$vote->setPixel(2*$e + 1, 2*$n + 0, $idx2);
				$vote->setPixel(2*$e + 1, 2*$n + 1, $idx2);
			}
			else {
				if ((($e & 3) == 2) && ($idx == 2 || $idx == 3)) {	# green stripe
					$idx = 6;
				}
				for (my $j = 0; $j < $SCALE; $j++) {
					for (my $i = 0; $i < $SCALE; $i++) {
						$vote->setPixel($SCALE*$e + $i, $SCALE*$n + $j, $idx);
					}
				}
			}
			#printf(STDERR $idx);
		}
		$count+=$percent_per_line;
		while ($count >= $percent) {
			printf(STDERR "%2d\r", ++$percent);
		}
	}

	open G,">",$gifname or die "$!";
	binmode G;
	print G $vote->gif;
	close(G);

	$world = substr($gifname,0,length($gifname)-3).'gfw';
	open(W, ">", $world) or die "$!";
	my $SCALE = 1;
	$X0 *= $SCALE;
	$Y1 = ($Y1+1)*$SCALE-1;
	foreach (1, 0, 0, -1, 
			 ($mine+0.5), 
			 ($maxn+0.5)
			) {
	  printf(W "%1.1f\n", $_);
	}
	close(W);
	return $vote;
}

sub deltas
{
	my @steps = sort {$a <=> $b} @_;
	return (undef, undef) if (scalar(@steps) <= 1);

	my ($min, $max) = (1e38, 0);
	for (my $i = 1; $i < scalar(@steps); $i++) {
		my $step = $steps[$i]-$steps[$i-1];
		$min = $step if ($step < $min);
		$max = $step if ($step > $max);
	}
	return ($min, $max);
}

sub find_min_max
{
	my @files = @_;
	
	my ($xmin, $xmax, $ymin, $ymax) = (1e38,-1e38,1e38,-1e38);
	if (scalar(@files) > 1) {
		# Try to use fast method: Look for tile_EEEEEE_NNNNNNN names!
		my %ehist = ();
		my %nhist = ();
		my $fast = 1;
		foreach (@files) {
			unless (/tile_([\-]?\d+)_([\-]?\d+)\D/) {
				$fast = 0;
				last;
			}
			$ehist{$1}++;
			$nhist{$2}++;
		}
		if ($fast) {
			my @ehist = sort {$a <=> $b} keys %ehist;
			my @nhist = sort {$a <=> $b} keys %nhist;
			my ($min, $max) = deltas(@ehist);
			my $step = $min;
			($min, $max) = deltas(@nhist);
			if ($min == $step) {
				$step = 250 if ($step < 250);
				
				$xmin = $ehist[0];
				$xmax = $ehist[-1] + $step;
				$ymin = $nhist[0];
				$ymax = $nhist[-1] + $step;
					
				return ($xmin, $xmax, $ymin, $ymax, $step);
			}
		}
	}
	my $ground_points = 0;
	foreach (@files) {
		chomp;
		my $path = $_;
		$path =~ s/\//\\/g;
		my $res = `lasinfo -stdout $path`;
		if (0) { #($res =~ /LAStiling .+ bbox (\S+)\s(\S+)\s(\S+)\s+(\S+), buffer/) {
			$xmin = $1 if ($1 < $xmin);
			$ymin = $2 if ($2 < $ymin);
			$xmax = $3 if ($3 > $xmax);
			$ymax = $4 if ($4 > $ymax);
		}
		else {
			die("Not a LAS/LAZ file? $_\n") unless ($res =~ /min x y z:\s+(\S+)\s+(\S+)\s+(\S+)/);
			$xmin = $1 if ($1 < $xmin);
			$ymin = $2 if ($2 < $ymin);
			die("Not a LAS/LAZ file? $_\n") unless ($res =~ /max x y z:\s+(\S+)\s+(\S+)\s+(\S+)/);
			$xmax = $1 if ($1 > $xmax);
			$ymax = $2 if ($2 > $ymax);
		}
		my $g = 0;
		if ($res =~ /\s(\d+)\s+Ground (2)/) {
			$g = $1;
		}
		$ground_points += $g;
	}
	#printf(STDERR "Raw X: %1.2f - %1.2f\nY: %1.2f - %1.2f\n", $xmin, $xmax, $ymin, $ymax);
	
	$xmin = adjust($xmin, 50);
	$ymin = adjust($ymin, 50);
	$xmax = adjust($xmax, -50);
	$ymax = adjust($ymax, -50);
	
	return ($xmin, $xmax, $ymin, $ymax);
}

sub adjust
{
	my ($x, $adj) = @_;
	$x = int($x+0.5); 		# Round to nearest integer!
	my $ax = $x+$adj; 		# Round to nearest integer!
	$x = $ax if (($ax % 100) == 0);
	return $x;
}

sub myglob
{
	my @match = ();
	foreach (@_) {
		my $dir = '.';
		if (/^(.+)[\\\/]([^\\\/]+)$/) { # Starts with a directory!
			$dir = $1;
			$_ = $2;
		}
		s/\./\t/g;		# Save away any explicit '.' symbols
		s/\*/\.*/g;		# Convert * to .*
		s/\?/\./g;		# Convert ? to .
		s/\t/\\\./g;	# Restore explicit '.' as escaped '\.'
		
		my $pattern = $_;
		
		opendir(D, $dir);
		my @files = readdir(D);
		closedir(D);
		
		$dir = ($dir eq '.') ? '' : $dir .'/';
		
		foreach (@files) {
			next unless (-f $dir.$_);
			
			next unless (/^$pattern$/);
			push(@match, $dir . $_);
		}
	}
	return @match;
}

