#!perl -w

use strict;
use Cwd;
use GD;

# Updated 2018-12-17 with a patch from Sverre Skoyen for better OOM compatibility

my $BORDERS = 0;
my $VERBOSE = 0;
my $MERGE = 0;
my ($gifname, $dxf);

my $CORES = 0; # If > 1, run separate threads for the tile processing

while ($_ = shift) {
	if (/^-borders$/i) {
		$BORDERS = 1;            # Generate lines for all border segments
	}
	elsif (/^-cores$/i) {
		$CORES = shift;
		if (! ($CORES =~ /^\d+$/)) {
			$CORES = 0;
		}
	}
	elsif (/^-verbose$/i) {
		$VERBOSE = 1;
	}
	elsif (/^-merge$/i) {
		$MERGE = 1;
	}
	elsif (/^-/) {
		# NOP
	}
	elsif (!defined($gifname)) {
		$gifname = $_;
	}
	elsif (!defined($dxf)) {
		$dxf = $_;
	}
}
$CORES = 1 if ($CORES < 1);

my @files = myglob($gifname);

if (scalar(@files) == 1 && ! ($files[0] =~ /\.gif$/i)) { # single non-GIF file, use as List of Files:
	open(LOF, $files[0]);
	@files = ();
	while (<LOF>) {
		chomp;
		push(@files, $_);
	}
	close(LOF);
}

die("Usage: veg2dxf.pl giffile|file_with_list_of_gif_files\n") unless (scalar(@files));

my %junctions;
my @lines;
my @freeLines;

my $ERASED;

my $CID = '';
my @tids = ();

if ($CORES > 1) {
	for (my $core = 0; $core < $CORES; $core++) {
		# Generate interleaved work file list:
		my @flist = ();
		for (my $f = $core; $f < scalar(@files); $f += $CORES) {
			push(@flist, $files[$f]);
		}
		my $tid = fork;
		if ($tid == 0) {
			$CID = sprintf("%d: ", $core);
			my $f = 1;
			foreach(@flist) {
				printf(STDERR "%s%s %d/%d\n", 
					$CID, $_, $f++, scalar(@flist));
				veg2dxf($_);
			}
			die("Core $core done!\n");
		}
		else {
			printf(STDERR "PID $tid forked\n");
			push(@tids, $tid);
		}
	}
}
else {
	foreach(@files) {
		my $f = 1;
		printf(STDERR "%s%s %d/%d\n", 
			$CID, $_, $f++, scalar(@files));
		veg2dxf($_);
	}
}

if ($MERGE) {
	foreach (@tids) {
		my $kid = waitpid($_, 0);
	}
	printf(STDERR "Finished waiting, merging individual vegetation files\n");
	
	my @body = ();
	my ($xmin, $ymin, $xmax, $ymax) = (1e38,1e38,-1e38,-1e38);
	my $i = 0;
	foreach (@files) {
		my $fname = substr($_, 0, -3).'dxf';
		printf(STDERR "%d/%d %s\n", ++$i, scalar(@files), $fname);
		open(F, $fname) || die("Could not open $fname\n");
		my @file = (<F>);
		close(F);
		for (my $i = 0; $i < 22; $i++) {
			if ($file[$i] =~ /\$EXTMIN/i) {
				my $e = $file[$i+2] + 0;
				my $n = $file[$i+4] + 0;
				$xmin = $e if ($e < $xmin);
				$ymin = $n if ($n < $ymin);
			}
			elsif ($file[$i] =~ /\$EXTMAX/i) {
				my $e = $file[$i+2] + 0;
				my $n = $file[$i+4] + 0;
				$xmax = $e if ($e > $xmax);
				$ymax = $n if ($n > $ymax);
			}
		}
		#printf(STDERR " (%s %s) -> (%s %s)\n", $xmin, $ymin, $xmax, $ymax);
		my @b = (@file)[22..scalar(@file)-5];
		push(@body, @b);
	}
	if (! defined($dxf)) {
		$dxf = getcwd;
		if ($dxf =~ /^(.+)[\\\/]([^\\\/]+)$/) {
			$dxf = $2;
		}
		$dxf = $dxf . '_vegetation.dxf';
	}
	else {
		$dxf = $dxf . '.dxf' unless ($dxf =~ /\.dxf$/);
	}
	open (F, '>', $dxf) || die("Could not write to $dxf\n");
	printf(F "0\nSECTION\n2\nHEADER\n9\n\$EXTMIN\n10\n%1.2f\n20\n%1.2f\n9\n\$EXTMAX\n10\n%1.2f\n20\n%1.2f\n0\nENDSEC\n0\nSECTION\n2\nENTITIES\n",
		$xmin, $ymin, $xmax, $ymax);
	foreach (@body) {
		print(F);
	}
	printf(F "0\nENDSEC\n0\nEOF\n");
	close(F);
	
	my $crt = substr($dxf, 0, length($dxf)-3).'crt';
	open(F, '>', $crt) || die("Could not create CRT file: $crt\n");
	print(F 
q(211 000_white
999 999_black
407 407_green_stripe
409 409_green_stripe
413 403_407_green_yellow_stripe
403 403_yellow
406 406_light_green
408 408_green
410 410_dark_green
));
	close(F);
	printf(STDERR "All done!\n");
	
	
}

#veg2dxf($gifname, $dxfname);
##############################################
#die("Done!\n");
##############################################

sub jkey
{
	my ($x, $y) = @_;
	if (!defined($y)) {
		($x, $y) = split(/,/, $x);
	}
	return sprintf("%d,%d", $x, $y);
}

sub saveSegment
{
	my @coords = @_;
	die("Too few points (< 2)!") if (scalar(@coords) < 2);
	
	my ($head, $tail) = (jkey($coords[0]), jkey($coords[-1]));
	my $idx = shift @freeLines;
	if (!defined($idx)) {
		$idx = scalar(@lines);
		push(@lines,'');
	}
	$lines[$idx] = join(";", @coords);
	push(@{$junctions{$head}}, $idx);
	push(@{$junctions{$tail}}, $idx);
}

sub freeSegment
{
	my ($idx) = @_;
	my @coords = split(/;/, $lines[$idx]);
	die("Too few points (< 2)!") if (scalar(@coords) < 2);
	if (scalar(@coords) >= 2) {
		my ($head, $tail) = (jkey($coords[0]), jkey($coords[-1]));
		foreach ($head, $tail) {
			my $k = $_;
			my @j = @{$junctions{$k}};
			my @n = ();
			foreach (@j) {
				push(@n, $_) unless ($_ == $idx);
			}
			@{$junctions{$k}} = @n;
		}
		push(@freeLines, $idx);
	}
	$lines[$idx] = '';
}
	
sub save
{
	my ($x0,$y0, $x1,$y1) = @_;
	saveSegment(sprintf("%d,%d", $x0,$y0), sprintf("%d,%d", $x1,$y1));
}

sub checkLines
{
	printf(STDERR "%sChecking", $CID) if ($VERBOSE);
	my %seen_junc = ();
	foreach (@lines) {
		next unless ($_);
		my @coords = split(/;/);
		die("Too few points (< 2)!") if (scalar(@coords) < 2);
		my ($head, $tail) = (jkey($coords[0]), jkey($coords[-1]));
		# Check that both endpoints are indexed:
		die("Missing junction(s)\n") unless (defined($junctions{$head}) && defined($junctions{$tail}));
		$seen_junc{$head}++;
		$seen_junc{$tail}++;
	}
	foreach (keys %junctions) {
		next unless ($_);
		next unless (scalar(@{$junctions{$_}}));
		next if ($seen_junc{$_});
		
		printf(STDERR "%sJunction %s is not part of any lines\n", $CID, $_);
		die;
	}
}

sub joinLines
{
	my ($sizex, $sizey) = @_;
	my $merged;
	# Insert dummy line segments pointing into each outside corner:
	foreach (0, $sizey) {
		my $y = $_;
		foreach (0,$sizex) {
			push(@{$junctions{"$_,$y"}}, -1);
		}
	}
	my $l = sprintf("%sJoining line segments: Going from %d to ", $CID, scalar(@lines)-scalar(@freeLines));
	do {
		$merged = 0;
		foreach (keys %junctions) {
			my @l = @{$junctions{$_}};
			if (scalar(@l) == 2 && ($l[0] != $l[1])) {  # Not a loop!
				
				# Remove two line segments, then insert the merged one!
				my @c0 = split(/;/, $lines[$l[0]]);
				freeSegment($l[0]);
				my @c1 = split(/;/, $lines[$l[1]]);
				freeSegment($l[1]);
				
				my $head0 = jkey($c0[0]);
				my $tail0 = jkey($c0[-1]);
				my $head1 = jkey($c1[0]);
				my $tail1 = jkey($c1[-1]);
				
				my @c;
				if ($tail0 eq $head1) {
					shift @c1;
					@c = (@c0, @c1);
				}
				elsif ($tail0 eq $tail1) {
					@c1 = reverse @c1;
					shift @c1;
					@c = (@c0, @c1);
				}
				elsif ($head0 eq $head1) {
					@c0 = reverse @c0;
					shift @c1;
					@c = (@c0, @c1);
				}
				elsif ($head0 eq $tail1) {
					shift @c0;
					@c = (@c1, @c0);
				}
				else {
					die("Junction error!\n");
				}
				saveSegment(@c);
				$merged++;
			}
		}
	} while ($merged);
	#checkLines;
	
	# Compact and re-index all the filtered lines!
	my @ln = @lines;
	@lines = ();
	@freeLines = ();
	%junctions = ();
	foreach (@ln) {
		my @c = split(/;/);
		next unless (scalar(@c) >= 2);
		saveSegment(@c);
	}
	#checkLines;
	printf(STDERR "%s%d lines\n", $l, scalar(@lines)-scalar(@freeLines)) if ($VERBOSE);
}

my @startPoints = ();

sub extractVegetationClasses
{
	my ($gf, $sizex, $sizey) = @_;
	# Get rid of stripes by copying the second pixel on top of the first:
	for (my $y = 0; $y < $sizey; $y+=1) {
		for (my $x = 0; $x < $sizex; $x+=2) {
			$gf->setPixel($x, $y, $gf->getPixel($x+1, $y));
		}
	}
	# Scan (only checking the top-left pixel in each 2x2 quad) for any remaining non-erased pixels:
	for (my $y = 1; $y < $sizey; $y+=2) {
		for (my $x = 1; $x < $sizex; $x+=2) {
			my $curr = $gf->getPixel($x, $y);
			if ($curr != $ERASED) { # Any kind of non-erased pixel!
				push(@startPoints, sprintf("%d,%d,%d", $x, $y, $curr));
				# Flood fill with $ERASED, i.e. RED:
				$gf->fill($x,$y,$ERASED);
			}
		}
	}
}

sub filterLines
{
	my ($icnt, $ocnt) = (0,0);
	my @l = ();
	for (my $l = 0; $l < scalar(@lines); $l++) {
		my @coords = split(/;/, $lines[$l]);
		$icnt += scalar(@coords);
		next if (scalar(@coords) < 2);

		# First filter pass: Make 45-degree bends in every 90-degree corner, 
		# skip intermediate points for horizontal/vertical lines:
		my @c = ($coords[0]); # Always keep the starting point
		my ($ppx, $ppy) = split(/,/, $coords[0]);
		my ($px, $py) = split(/,/, $coords[1]);

		for (my $i = 2; $i < scalar(@coords); $i++) {
			my ($x, $y) = split(/,/, $coords[$i]);
			if (($ppx-$px != $px-$x) || ($ppy-$py != $py-$y)) { # Not a straight line, insert bend:
				push(@c, sprintf("%1.1f,%1.1f",($ppx+$px)*0.5,($ppy+$py)*0.5));
				push(@c, sprintf("%1.1f,%1.1f",($px+$x)*0.5,($py+$y)*0.5));
			}
			($ppx,$ppy) = ($px, $py);
			($px,$py) = ($x, $y);
		}
		push(@c, $coords[-1]); # Keep end point!
		@coords = @c;
		
		# Second filter pass, skip intermediate points in 45-degree staircase lines:
		@c = ($coords[0]);
		($ppx, $ppy) = split(/,/, $coords[0]);
		($px, $py) = split(/,/, $coords[1]);

		for (my $i = 2; $i < scalar(@coords); $i++) {
			my ($x, $y) = split(/,/, $coords[$i]);
			if (($ppx-$px != $px-$x) || ($ppy-$py != $py-$y)) { # Not a straight line, keep point:
				push(@c, join(",", $px, $py));
			}
			($ppx,$ppy) = ($px, $py);
			($px,$py) = ($x, $y);
		}
		push(@c, $coords[-1]);
		
		$ocnt += scalar(@c);
		push(@l, join(";", @c));
	}
	@lines = @l;
	printf(STDERR "%sStarted with %d points, %d after filtering\n", $CID, $icnt, $ocnt) if ($VERBOSE);
}

sub writeLines
{
	my ($dxf, $xtopleft, $ytopleft) = @_;
	my $a=0;
	foreach (@lines) {
		my @coords = split(/;/);
		next if (scalar(@coords) < 2);
		#index_coordinates($a, @coords);
		
		printf($dxf "0\nLWPOLYLINE\n8\n%s\n%s90\n%d\n", 'vegborder', '', scalar(@coords));
		foreach (@coords) {
			my ($x, $y) = split(/,/);
			printf( $dxf "10\n%1.2f\n20\n%1.2f\n", $xtopleft+$x, $ytopleft-$y);
		}
		printf( $dxf "0\nSEQEND\n");
	}
}

sub writeStartpoints
{
	my ($dxf, $xtopleft, $ytopleft) = @_;
	my @vegClasses = ('white', 'black', 'green_white_stripe', 'green_yellow_stripe', 'yellow', 'light_green', 'green', 'dark_green');
	
	foreach (@startPoints) {
		my ($x, $y, $veg) = split(/,/);
		
		printf( $dxf "0\nPOINT\n8\n%s\n10\n%1.2f\n20\n%1.2f\n%s0\nSEQEND\n", $vegClasses[$veg], $xtopleft+$x, $ytopleft-$y, '');
	}
	printf(STDERR "%sWrote %d area starting points\n", $CID, scalar(@startPoints)) if (scalar(@startPoints) && $VERBOSE);
}

sub find_area_min_max
{
	my (@boundary) = @_;
	my ($xmin, $ymin, $xmax, $ymax) = (1e38,1e38,-1e38,-1e38);
	foreach (@boundary) {
		my ($x, $y, $h) = split(/,/);
		$xmin = $x if ($x < $xmin);
		$ymin = $y if ($y < $ymin);
		$xmax = $x if ($x > $xmax);
		$ymax = $y if ($y > $ymax);
	}
	return ($xmin, $ymin, $xmax, $ymax);
}

sub point_is_inside
{
	my ($sx,$sy, $boundary) = @_;
	return 0 if (scalar(@{$boundary}) < 3);
	
	my ($x, $y);
	my $crossings = 0;
	my $last_x_crossing = -1e38;
	foreach (@{$boundary}) {
		my ($px, $py) = ($x, $y);
		($x, $y) = split(/,/);
		next unless (defined($py));
		
		# Must cross the Y coordinate!
		next if ($y >= $sy && $py >= $sy);
		next if ($y < $sy && $py < $sy);
		
		# X intercept coord must be > starting point!
		next if ($px < $sx && $x < $sx);
		
		# Possible crossing to the right!
		my $ix = $px + ($x-$px) * ($sy-$py) / ($y - $py);
		if ($ix > $sx) {
			$crossings++;
			if ($ix > $last_x_crossing) {
				$last_x_crossing = $ix;
			}
		}
	}
	return ($crossings & 1, $last_x_crossing);
}

sub locateAreas
{
	my ($dxf, $xtopleft, $ytopleft) = @_;

	my @areas = ();
	my @sp = ();
	foreach (@startPoints) {
		my %area = ();
		my ($x, $y, $veg) = split(/,/);
		my @boundary = find_boundary($x, $y, \@lines);
		if (!scalar(@boundary)) {
			push(@sp, $_);
			next;
		}
		
		$area{x} = $x;
		$area{y} = $y;
		$area{veg} = $veg;
		($area{minx},$area{miny},$area{maxx},$area{maxy}) = find_area_min_max(@boundary);
		@{$area{boundary}} = @boundary;
		@{$area{inner}} = ();
		
		push(@areas, \%area);
	}
	@areas = sort {$a->{x} <=> $b->{x}} @areas;
	
	my @outer = ();
	for (my $i = 0; $i < scalar(@areas); $i++) {
		my $c = $areas[$i];
		my ($x0, $y0, $x1, $y1) = ($c->{minx},$c->{miny},$c->{maxx},$c->{maxy});
		my @boundary = @{$c->{boundary}};
		for (my $j = $i+1; $j < scalar(@areas); $j++) {
			my $n = $areas[$j];
			my ($xn0, $yn0, $xn1, $yn1) = ($n->{minx},$n->{miny},$n->{maxx},$n->{maxy});
			last if ($xn0 > $x1);
			next if ($xn1 > $x1 || $yn0 < $y0 || $yn1 > $y1);
			my ($inside, $last_x) = point_is_inside($n->{x}, $n->{y}, \@boundary);
			next unless $inside;
			$outer[$j] = $i;
		}
	}
	for (my $i = 0; $i < scalar(@areas); $i++) {
		if (defined($outer[$i])) {
			push(@{$areas[$outer[$i]]->{inner}}, $i);
			#printf(STDERR "Area %d contains area %d\n", $outer[$i], $i);
		}
	}
	my @vegClasses = ('000_white', '999_black', '409_green_stripe', "403_yellow;409_green_stripe", 
		'403_yellow', '406_light_green', '408_green', '410_dark_green');
	for (my $i = 0; $i < scalar(@areas); $i++) {
		my $c = $areas[$i];
		my $veg = $c->{veg};
		
		next if ($veg == 0); # Skip white areas!
		
		my @coords = @{$c->{boundary}};
		
		foreach (split(/;/,$vegClasses[$veg])) { # Handle any combined areas, i.e. yellow + green stripes!
			# Include 92 / 2 parameter from Sverre Foyen:
			printf( $dxf "  0\nHATCH\n 5\n27\n100\nAcDbEntity\n  8\n%s\n%s100\nAcDbEntity\n  2\nSOLID\n 70\n1\n 91\n2\n 92\n2\n 73\n1\n 93\n%d\n",
				$_, '', scalar(@coords));
			foreach (@coords) {
				my ($x, $y, $h) = split(/,/);
				printf( $dxf "10\n%1.2f\n20\n%1.2f\n", $xtopleft+$x, $ytopleft-$y);
			}

			foreach (@{$c->{inner}}) {
				my @points = @{$areas[$_]->{boundary}};
				printf( $dxf " 73\n1\n 93\n%d\n", scalar(@points));
				foreach (@points) {
					my ($x, $y, $h) = split(/,/);
					printf( $dxf "10\n%1.2f\n20\n%1.2f\n", $xtopleft+$x, $ytopleft-$y);
				}
			}
			printf( $dxf "0\nSEQEND\n");
		}
	}
	printf(STDERR "%sWrote %d area symbols (%d remaining)\n", $CID, scalar(@areas), scalar(@sp)) if ($VERBOSE);
	@startPoints = @sp;
}

sub veg2dxf
{
	my ($gifname,$dxfname) = @_;
	$dxfname = substr($gifname,0,length($gifname)-3).'dxf' unless (defined($dxfname));

	%junctions = ();
	@lines = @freeLines = ();
	
	printf(STDERR "%sReading $gifname\n", $CID) if ($VERBOSE);
	
	my $gf;
	my ($xtopleft, $ytopleft);
	open(my $gif, '<', $gifname);
	binmode($gif);
	$gf = GD::Image->newFromGif($gif);
	close($gif);
	#printf(STDERR " done!\n");
	my $gfwname = substr($gifname,0,length($gifname)-3).'gfw';
	#printf(STDERR "Reading $gfwname file...");
	open(my $gw, '<', $gfwname);
	#binmode($gw);
	my @world = <$gw>;
	close($gw);
	($xtopleft, $ytopleft) = ($world[4] - $world[0]/2,$world[5] - $world[3]/2);
	#printf(STDERR " done!\n");
	$ERASED = $gf->colorAllocate(255,0,0);
	my ($sizex, $sizey) = $gf->getBounds;

	open(my $dxf, '>', $dxfname);
	printf($dxf "0\nSECTION\n2\nHEADER\n9\n\$EXTMIN\n10\n%1.2f\n20\n%1.2f\n9\n\$EXTMAX\n10\n%1.2f\n20\n%1.2f\n0\nENDSEC\n",
			$xtopleft, $ytopleft-$sizey, $xtopleft+$sizex, $ytopleft);
	printf($dxf "0\nSECTION\n2\nENTITIES\n");
	
	for (my $y = 0; $y <= $sizey; $y+=2) {
		for (my $x = 0; $x <= $sizex; $x+=2) {

			my $curr = $gf->getPixel($x+1, $y);
			my $left = $gf->getPixel($x-1, $y);
			my $abov = $gf->getPixel($x+1, $y-2);
			# Vertical line left of this position?
			if (($left != $curr) || $x == 0 || $x >= $sizex) {
				#printf(STDERR "save($x, $y, $x, %d)", $y+2);
				save($x, $y, $x, $y+2);
			}
			# Horizontal line above this position?
			if (($abov != $curr || $y == 0 || $y >= $sizey)) {
				#printf(STDERR "save($x, $y, %d, $y)", $x+2);
				save($x, $y, $x+2, $y);
			}
		}
	}
	for (my $y = 0; $y < $sizey; $y+=2) {
	#	save(0, $y, 0, $y+2);
	#	save($sizex, $y, $sizex, $y+2);
	}

	joinLines($sizex, $sizey);
	#checkLines;

	printf(STDERR "%sExtracting starting points for areas\n", $CID) if ($VERBOSE);
	extractVegetationClasses($gf, $sizex, $sizey);
	printf(STDERR "%sFiltering the lines\n", $CID) if ($VERBOSE);
	filterLines;
	#checkLines;

	if ($BORDERS) {
		printf(STDERR "%sWriting the area border lines\n", $CID) if ($VERBOSE);
		writeLines($dxf, $xtopleft, $ytopleft);
		#checkLines;
	}

	printf(STDERR "%sLocating areas\n", $CID) if ($VERBOSE);
	locateAreas($dxf, $xtopleft, $ytopleft);
	#printf(STDERR "Writing any remaining starting points\n");
	writeStartpoints($dxf, $xtopleft, $ytopleft);

	printf( $dxf "0\nENDSEC\n0\nEOF\n");
	close($dxf);
}

sub find_boundary
{
	my ($x, $y, $arg) = @_;

	my ($inside, $last_x);
	my @boundary;
	do {
		# Find nearest area line segment:
		my ($start, $startdown) = nearest_to_the_right($x, $y, $arg);
		#printf(STDERR "Starting at (%d,%d), nearest line not found!\n", $x, $y) unless(defined($start));
		return () unless(defined($start));

		# Search for the shortest path around the ARF flood fill point:
		# We're starting with $arg->[$start], now we must locate the junctions:
		my @coords = split(/;/, $arg->[$start]);
		if (! $startdown) {# Reverse them, so that we go clockwise around!
			@coords = reverse @coords;
		}
		my $curr = $start;
		my ($jkey0, $jkey1) = (jkey(split(/,/,$coords[0])), jkey(split(/,/,$coords[-1])));
		if (!defined($junctions{$jkey1})) {
			printf(STDERR "Listing line segment endpoints\n");
			my $a = 0;
			foreach (@$arg) {
				my @coords = split(/;/);
				my $jkey = jkey(split(/,/,$coords[0]));
				if (defined($junctions{$jkey})) {
					printf(STDERR "Line $a starts at junction %d\n", $junctions{$jkey});
				}
				else {
					printf(STDERR "Line $a start junction %s not found!\n", $jkey);
				}
				$a++;
			}
			die("Junction $jkey1 not found!\n") 
		}
		#printf(STDERR "Starting at ($x,$y), first line to the right is $start. %d segments at first junction\n", scalar(@{$junctions{$jkey1}}));

		my %segments = ($start => 1);
		my %junc = ($jkey1 => 1);

		@boundary = @coords;

		my $nextangle;
		my @next;
		my $next;
		while ($jkey1 ne $jkey0) {
			$nextangle = 200; # Inf
			# Calculate incoming direction, from last two points
			my $idir = direction($boundary[-2],$boundary[-1]);
			#printf(STDERR "idir = %6.2f\n", $idir);
			
			foreach (@{$junctions{$jkey1}}) {
				#printf(STDERR "Trying segment %d\n", $_);
				
				next if (defined($segments{$_})); # Never use the same segment again!
				
				# Here is another border segment which starts in the same junction:
				@coords = split(/;/, $arg->[$_]);
				my $jkey = jkey(split(/,/,$coords[0]));
				if ($jkey ne $jkey1) {
					@coords = reverse @coords;
					$jkey = jkey(split(/,/,$coords[0]));
					die("Could not locate a matching segment!") if ($jkey ne $jkey1);
				}
				my $odir = direction($coords[0],$coords[1]);
				#printf(STDERR "  odir = %6.2f\n", $odir);
				my $angle = $odir - $idir;
				if ($angle < -180) {
					$angle += 360;
				}
				if ($angle > 180) {
					$angle -= 360;
				}
				#printf(STDERR "Angle = %6.2f\n", $angle);
				# Look for largest possible right turn => large negative angle:
				if ($angle < $nextangle) {
					$nextangle = $angle;
					$next = $_;
					@next = @coords;
				}
				# Is the current segment a loop?
				# If so we must check that we traverse in the correct direction!
				if (jkey(split(/,/,$coords[0])) eq jkey(split(/,/,$coords[-1]))) {
					#printf(STDERR "Found a loop, checking reverse direction...\n");
					@coords = reverse @coords;
					my $odir = direction($coords[0],$coords[1]);
					#printf(STDERR "  odir = %6.2f\n", $odir);
					my $angle = $odir - $idir;
					if ($angle < -180) {
						$angle += 360;
					}
					if ($angle > 180) {
						$angle -= 360;
					}
					#printf(STDERR "Angle = %6.2f\n", $angle);
					# Look for largest possible right turn => large negative angle:
					if ($angle < $nextangle) {
						$nextangle = $angle;
						$next = $_;
						@next = @coords;
						#printf(STDERR "Reverse was better!\n");
					}
				}
			}
			#printf(STDERR "next angle = %6.2f\n", $nextangle);
			@coords = @next;
			$segments{$next}++;
			shift @next;
			push(@boundary, @next);
			$jkey1 = jkey(split(/,/,$boundary[-1]));
			if ($nextangle > 180) {
				@boundary = ();
				printf(STDERR "Error locating next segment for $jkey1, skipping...\n");
				die("");
				return();
			}
		}
		return () if (scalar(@boundary) == 0);
		
		($inside, $last_x) = point_is_inside($x, $y, \@boundary);
		if (! $inside) {
			$x = $last_x + 0.01;
			printf(STDERR "Skipping internal area\n") if ($VERBOSE);
		}
	} while (! $inside);  # Scan past any embedded areas!
	
	return @boundary;
}

sub direction
{
	my ($c0, $c1) = @_;
	my ($x0, $y0) = split(/,/, $c0);
	my ($x1, $y1) = split(/,/, $c1);
	my ($dx, $dy) = ($x1-$x0, $y1-$y0);
	return atan2($dy,$dx)*(180/3.141592653589);
}

sub nearest_to_the_right
{
	my ($sx, $sy, $lines) = @_;
	
	my $bestx = 1e38;
	my $besta;
	my $bestdown;
	for (my $i = 0; $i < scalar(@{$lines}); $i++) {
		my @coords = split(/;/, $lines->[$i]);
		next unless (scalar(@coords) > 1);
		
		my ($px, $py, $ph) = split(/,/, shift @coords);
		foreach (@coords) {
			my ($x, $y, $h) = split(/,/);
			if ((($py <= $sy) && ($y > $sy)) || 
				(($py > $sy) && ($y <= $sy))) { # Passing the SY coordinate
				my ($ix, $iy) = ($px + ($x-$px)*($sy-$py)/($y-$py), $sy); # Intercept coordinates, ($y-$py) != 0
				if (($ix >= $sx) && ($ix < $bestx)) {
					$bestx = $ix;
					$besta = $i;
					$bestdown = $py > $y; 
				}
			}
			($px, $py) = ($x, $y);
		}
	}
	return ($besta, $bestdown);
}

sub myglob
{
	my @match = ();
	foreach (@_) {
		my $dir = '.';
		if (/^(.+)[\\\/]([^\\\/]+)$/) { # Starts with a directory!
			$dir = $1;
			$_ = $2;
		}
		s/\./\t/g;		# Save away any explicit '.' symbols
		s/\*/\.*/g;		# Convert * to .*
		s/\?/\./g;		# Convert ? to .
		s/\t/\\\./g;	# Restore explicit '.' as escaped '\.'
		
		my $pattern = $_;
		
		opendir(D, $dir);
		my @files = readdir(D);
		closedir(D);
		
		$dir = ($dir eq '.') ? '' : $dir .'/';
		
		foreach (@files) {
			next unless (-f $dir.$_);
			
			next unless (/^$pattern$/);
			push(@match, $dir . $_);
		}
	}
	return @match;
}

